# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""A serializer to emit YAML but with request body in nicely formatted JSON."""

import json
import os

import betamax.serializers.base
import yaml


def _should_use_block(value):
    for c in "\u000a\u000d\u001c\u001d\u001e\u0085\u2028\u2029":
        if c in value:
            return True
    return False


def _represent_scalar(self, tag, value, style=None):
    if style is None:
        if _should_use_block(value):
            style = '|'
        else:
            style = self.default_style

    node = yaml.representer.ScalarNode(tag, value, style=style)
    if self.alias_key is not None:
        self.represented_objects[self.alias_key] = node
    return node


def _unicode_representer(dumper, uni):
    node = yaml.ScalarNode(tag='tag:yaml.org,2002:str', value=uni)
    return node


def _indent_json(val):
    if not val:
        return ''

    return json.dumps(
        json.loads(val),
        indent=2,
        separators=(',', ': '),
        sort_keys=False,
        default=str,
    )


def _is_json_body(interaction):
    content_type = interaction['headers'].get('Content-Type', [])
    return 'application/json' in content_type


class YamlJsonSerializer(betamax.serializers.base.BaseSerializer):
    name = "yamljson"

    @staticmethod
    def generate_cassette_name(cassette_library_dir, cassette_name):
        return os.path.join(cassette_library_dir, f"{cassette_name}.yaml")

    def serialize(self, cassette_data):
        # Reserialize internal json with indentation
        for interaction in cassette_data['http_interactions']:
            for key in ('request', 'response'):
                if _is_json_body(interaction[key]):
                    interaction[key]['body']['string'] = _indent_json(
                        interaction[key]['body']['string']
                    )

        class MyDumper(yaml.Dumper):
            """Specialized Dumper which does nice blocks and unicode."""

        yaml.representer.BaseRepresenter.represent_scalar = _represent_scalar  # type: ignore[method-assign]

        MyDumper.add_representer(str, _unicode_representer)

        return yaml.dump(
            cassette_data, Dumper=MyDumper, default_flow_style=False
        )

    def deserialize(self, cassette_data):
        try:
            deserialized = yaml.safe_load(cassette_data)
        except yaml.error.YAMLError:
            deserialized = None

        if deserialized is not None:
            return deserialized
        return {}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import inspect

from docutils import nodes
from docutils.parsers import rst
from docutils.parsers.rst import directives
from docutils.statemachine import ViewList
from sphinx.util import logging
from sphinx.util.nodes import nested_parse_with_titles
from stevedore import extension

LOG = logging.getLogger(__name__)


class ListAuthPluginsDirective(rst.Directive):
    """Present a simple list of the plugins in a namespace."""

    option_spec = {
        'class': directives.class_option,
        'overline-style': directives.single_char_or_unicode,
        'underline-style': directives.single_char_or_unicode,
    }

    has_content = True

    def report_load_failure(mgr, ep, err):
        LOG.warning(f'Failed to load {ep.module_name}: {err}')

    def display_plugin(self, ext):
        overline_style = self.options.get('overline-style', '')
        underline_style = self.options.get('underline-style', '=')

        if overline_style:
            yield overline_style * len(ext.name)

        yield ext.name

        if underline_style:
            yield underline_style * len(ext.name)

        yield "\n"

        doc = inspect.getdoc(ext.obj)
        if doc:
            yield doc
            yield "\n"
            yield "------"
            yield "\n"

        for opt in sorted(ext.obj.get_options(), key=lambda opt: opt.name):
            summary = f":{opt.name}: {opt.help}"
            if opt.required:
                summary += " **(mandatory)**"
            yield summary
            yield "\n\n"
            yield "     :CLI options: {}\n".format(
                ', '.join([f'``{x}``' for x in opt.argparse_args])
            )
            yield "     :Environment variables: {}\n".format(
                ', '.join([f'``{x}``' for x in opt.argparse_envvars])
            )

        yield "\n"

    def run(self):
        mgr = extension.ExtensionManager(
            'keystoneauth1.plugin',
            on_load_failure_callback=self.report_load_failure,
            invoke_on_load=True,
        )

        result = ViewList()

        for name in sorted(mgr.names()):
            for lines in self.display_plugin(mgr[name]):
                for line in lines.splitlines():
                    ep = mgr[name]
                    try:
                        module_name = ep.entry_point.module_name
                    except AttributeError:
                        try:
                            module_name = ep.entry_point.module
                        except AttributeError:
                            module_name = ep.entry_point.value
                    result.append(line, module_name)

        # Parse what we have into a new section.
        node = nodes.section()
        node.document = self.state.document
        nested_parse_with_titles(self.state, result, node)

        return node.children


def setup(app):
    LOG.info('loading keystoneauth1 plugins')
    app.add_directive('list-auth-plugins', ListAuthPluginsDirective)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from oslo_config import fixture as config

from keystoneauth1 import fixture
from keystoneauth1 import loading
from keystoneauth1 import session
from keystoneauth1.tests.unit import utils


class FixturesTests(utils.TestCase):
    GROUP = uuid.uuid4().hex
    AUTH_TYPE = uuid.uuid4().hex

    def setUp(self):
        super().setUp()
        self.conf_fixture = self.useFixture(config.Config())

        # conf loading will still try to read the auth_type from the config
        # object and pass that to the get_plugin_loader method. This value will
        # typically be ignored and the fake plugin returned regardless of name
        # but it could be a useful differentiator and it also ensures that the
        # application has called register_auth_conf_options before simply
        # returning a fake plugin.
        loading.register_auth_conf_options(
            self.conf_fixture.conf, group=self.GROUP
        )

        self.conf_fixture.config(auth_type=self.AUTH_TYPE, group=self.GROUP)

    def useLoadingFixture(self, **kwargs):
        return self.useFixture(fixture.LoadingFixture(**kwargs))

    def test_endpoint_resolve(self):
        endpoint = "http://%(service_type)s/%(version)s/%(interface)s"
        loader = self.useLoadingFixture(endpoint=endpoint)

        endpoint_filter = {
            'service_type': 'compute',
            'service_name': 'nova',
            'version': (2, 1),
            'interface': 'public',
        }

        auth = loading.load_auth_from_conf_options(
            self.conf_fixture.conf, self.GROUP
        )
        sess = session.Session(auth=auth)

        loader_endpoint = loader.get_endpoint(**endpoint_filter)
        plugin_endpoint = sess.get_endpoint(**endpoint_filter)

        self.assertEqual("http://compute/2.1/public", loader_endpoint)
        self.assertEqual(loader_endpoint, plugin_endpoint)

    def test_conf_loaded(self):
        token = uuid.uuid4().hex
        endpoint_filter = {
            'service_type': 'compute',
            'service_name': 'nova',
            'version': (2, 1),
        }

        loader = self.useLoadingFixture(token=token)

        url = loader.get_endpoint('/path', **endpoint_filter)

        m = self.requests_mock.get(url)

        auth = loading.load_auth_from_conf_options(
            self.conf_fixture.conf, self.GROUP
        )
        sess = session.Session(auth=auth)
        self.assertEqual(self.AUTH_TYPE, auth.auth_type)

        sess.get('/path', endpoint_filter=endpoint_filter)

        self.assertTrue(m.called_once)

        self.assertEqual(token, m.last_request.headers['X-Auth-Token'])
        self.assertEqual(loader.project_id, sess.get_project_id())
        self.assertEqual(loader.user_id, sess.get_user_id())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty

from keystoneauth1 import discover
from keystoneauth1.identity.generic import base
from keystoneauth1.identity import v2
from keystoneauth1.identity import v3
from keystoneauth1 import session as ks_session


class Token(base.BaseGenericPlugin):
    """Generic token auth plugin.

    :param string token: Token for authentication.
    """

    def __init__(
        self,
        auth_url: str,
        token: str,
        *,
        tenant_id: ty.Optional[str] = None,
        tenant_name: ty.Optional[str] = None,
        project_id: ty.Optional[str] = None,
        project_name: ty.Optional[str] = None,
        project_domain_id: ty.Optional[str] = None,
        project_domain_name: ty.Optional[str] = None,
        domain_id: ty.Optional[str] = None,
        domain_name: ty.Optional[str] = None,
        system_scope: ty.Optional[str] = None,
        trust_id: ty.Optional[str] = None,
        default_domain_id: ty.Optional[str] = None,
        default_domain_name: ty.Optional[str] = None,
        reauthenticate: bool = True,
    ):
        super().__init__(
            auth_url=auth_url,
            tenant_id=tenant_id,
            tenant_name=tenant_name,
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
            domain_id=domain_id,
            domain_name=domain_name,
            system_scope=system_scope,
            trust_id=trust_id,
            default_domain_id=default_domain_id,
            default_domain_name=default_domain_name,
            reauthenticate=reauthenticate,
        )

        self._token = token

    def create_plugin(
        self,
        session: ks_session.Session,
        version: discover._PARSED_VERSION_T,
        url: str,
        raw_status: ty.Optional[str] = None,
    ) -> ty.Union[None, v2.Token, v3.Token]:
        if discover.version_match((2,), version):
            return v2.Token(
                url,
                self._token,
                trust_id=self._trust_id,
                tenant_id=self._project_id,
                tenant_name=self._project_name,
                reauthenticate=self.reauthenticate,
            )

        if discover.version_match((3,), version):
            return v3.Token(
                url,
                self._token,
                trust_id=self._trust_id,
                system_scope=self._system_scope,
                project_id=self._project_id,
                project_name=self._project_name,
                project_domain_id=self.project_domain_id,
                project_domain_name=self.project_domain_name,
                domain_id=self._domain_id,
                domain_name=self._domain_name,
                reauthenticate=self.reauthenticate,
            )

        return None

    def get_cache_id_elements(self) -> dict[str, ty.Optional[str]]:
        return {
            'auth_url': self.auth_url,
            'project_id': self._project_id,
            'project_name': self._project_name,
            'project_domain_id': self.project_domain_id,
            'project_domain_name': self.project_domain_name,
            'domain_id': self._domain_id,
            'domain_name': self._domain_name,
            'trust_id': self._trust_id,
            'token': self._token,
        }



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

import betamax
from betamax import exceptions
import testtools

from keystoneauth1.fixture import keystoneauth_betamax
from keystoneauth1.fixture import serializer
from keystoneauth1.fixture import v2 as v2Fixtures
from keystoneauth1.identity import v2
from keystoneauth1 import session


class TestBetamaxFixture(testtools.TestCase):
    TEST_USERNAME = 'test_user_name'
    TEST_PASSWORD = 'test_password'
    TEST_TENANT_NAME = 'test_tenant_name'
    TEST_AUTH_URL = 'http://keystoneauth-betamax.test/v2.0/'

    V2_TOKEN = v2Fixtures.Token(
        tenant_name=TEST_TENANT_NAME, user_name=TEST_USERNAME
    )

    def setUp(self):
        super().setUp()
        self.ksa_betamax_fixture = self.useFixture(
            keystoneauth_betamax.BetamaxFixture(
                cassette_name='ksa_betamax_test_cassette',
                cassette_library_dir='keystoneauth1/tests/unit/data/',
                record=False,
            )
        )

    def _replay_cassette(self):
        plugin = v2.Password(
            auth_url=self.TEST_AUTH_URL,
            password=self.TEST_PASSWORD,
            username=self.TEST_USERNAME,
            tenant_name=self.TEST_TENANT_NAME,
        )
        s = session.Session()
        s.get_token(auth=plugin)

    def test_keystoneauth_betamax_fixture(self):
        self._replay_cassette()

    def test_replay_of_bad_url_fails(self):
        plugin = v2.Password(
            auth_url='http://invalid-auth-url/v2.0/',
            password=self.TEST_PASSWORD,
            username=self.TEST_USERNAME,
            tenant_name=self.TEST_TENANT_NAME,
        )
        s = session.Session()
        self.assertRaises(exceptions.BetamaxError, s.get_token, auth=plugin)


class TestBetamaxFixtureSerializerBehaviour(testtools.TestCase):
    """Test the fixture's logic, not its monkey-patching.

    The setUp method of our BetamaxFixture monkey-patches the function to
    construct a session. We don't need to test that particular bit of logic
    here so we do not need to call useFixture in our setUp method.
    """

    @mock.patch.object(betamax.Betamax, 'register_serializer')
    def test_can_pass_custom_serializer(self, register_serializer):
        serializer = mock.Mock()
        serializer.name = 'mocked-serializer'
        fixture = keystoneauth_betamax.BetamaxFixture(
            cassette_name='fake',
            cassette_library_dir='keystoneauth1/tests/unit/data',
            serializer=serializer,
        )

        register_serializer.assert_called_once_with(serializer)
        self.assertIs(serializer, fixture.serializer)
        self.assertEqual('mocked-serializer', fixture.serializer_name)

    def test_can_pass_serializer_name(self):
        fixture = keystoneauth_betamax.BetamaxFixture(
            cassette_name='fake',
            cassette_library_dir='keystoneauth1/tests/unit/data',
            serializer_name='json',
        )

        self.assertIsNone(fixture.serializer)
        self.assertEqual('json', fixture.serializer_name)

    def test_no_serializer_options_provided(self):
        fixture = keystoneauth_betamax.BetamaxFixture(
            cassette_name='fake',
            cassette_library_dir='keystoneauth1/tests/unit/data',
        )

        self.assertIs(serializer.YamlJsonSerializer, fixture.serializer)
        self.assertEqual('yamljson', fixture.serializer_name)

    def test_no_request_matchers_provided(self):
        fixture = keystoneauth_betamax.BetamaxFixture(
            cassette_name='fake',
            cassette_library_dir='keystoneauth1/tests/unit/data',
        )

        self.assertDictEqual({}, fixture.use_cassette_kwargs)

    def test_request_matchers(self):
        fixture = keystoneauth_betamax.BetamaxFixture(
            cassette_name='fake',
            cassette_library_dir='keystoneauth1/tests/unit/data',
            request_matchers=['method', 'uri', 'json-body'],
        )

        self.assertDictEqual(
            {'match_requests_on': ['method', 'uri', 'json-body']},
            fixture.use_cassette_kwargs,
        )



# keystoneauth Release Notes documentation build configuration file

# -- General configuration ------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['openstackdocstheme', 'reno.sphinxext']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'keystoneauth Release Notes'
copyright = '2015-, Keystone Developers'

# Release notes are version independent.

# The short X.Y version.

# The full version, including alpha/beta/rc tags.
release = ''
# The short X.Y version.
version = ''

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'native'


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'openstackdocs'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


# -- Options for LaTeX output ---------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        'index',
        'keystoneauthReleaseNotes.tex',
        'keystoneauth Release Notes Documentation',
        'Keystone Developers',
        'manual',
    )
]


# -- Options for Internationalization output ------------------------------
locale_dirs = ['locale/']


# -- Options for openstackdocstheme -------------------------------------------
openstackdocs_repo_name = 'openstack/keystoneauth'
openstackdocs_auto_name = False
openstackdocs_bug_project = 'keystoneauth'
openstackdocs_bug_tag = 'doc'



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

UNSCOPED_TOKEN = {
    "token": {
        "issued_at": "2014-06-09T09:48:59.643406Z",
        "extras": {},
        "methods": ["oidc"],
        "expires_at": "2014-06-09T10:48:59.643375Z",
        "user": {
            "OS-FEDERATION": {
                "identity_provider": {"id": "bluepages"},
                "protocol": {"id": "oidc"},
                "groups": [{"id": "1764fa5cf69a49a4918131de5ce4af9a"}],
            },
            "id": "oidc_user%40example.com",
            "name": "oidc_user@example.com",
        },
    }
}

ACCESS_TOKEN_VIA_PASSWORD_RESP = {
    "access_token": "z5H1ITZLlJVDHQXqJun",
    "token_type": "bearer",
    "expires_in": 3599,
    "scope": "openid profile",
    "refresh_token": "DCERsh83IAhu9bhavrp",
}

ACCESS_TOKEN_VIA_AUTH_GRANT_RESP = {
    "access_token": "ya29.jgGIjfVrBPWLStWSU3eh8ioE6hG06QQ",
    "token_type": "Bearer",
    "expires_in": 3600,
    "refresh_token": "1/ySXNO9XISBMIgOrJDtdun6zK6XiATCKT",
    "id_token": "eyJhbGciOiJSUzI1Ni8hOYHuZT8dt_yynmJVhcU",
}

DEVICE_CODE_RESP = {
    'device_code': "fRtNXLxmRy4iEzLsbAV96pMFgOMEUCPZmAMepjLrvQ",
    'user_code': "ABCD-EFGH",
    'verification_uri': "https://localhost:8020/oidc/authorize/device",
    'verification_uri_complete': "https://localhost:8020/oidc/authorize/device?user_code=ABCD-EFGH",
    'expires_in': 1800,
    'interval': 5,
}

DISCOVERY_DOCUMENT = {
    "authorization_endpoint": "https://localhost:8020/oidc/authorize",
    "claims_supported": [
        "sub",
        "name",
        "preferred_username",
        "given_name",
        "family_name",
        "middle_name",
        "nickname",
        "profile",
        "picture",
        "website",
        "gender",
        "zoneinfo",
        "locale",
        "updated_at",
        "birthdate",
        "email",
        "email_verified",
        "phone_number",
        "phone_number_verified",
        "address",
    ],
    "grant_types_supported": [
        "authorization_code",
        "password",
        "urn:ietf:params:oauth:grant-type:device_code",
    ],
    "introspection_endpoint": "https://localhost:8020/oidc/introspect",
    "issuer": "https://localhost:8020/oidc/",
    "jwks_uri": "https://localhost:8020/oidc/jwk",
    "op_policy_uri": "https://localhost:8020/oidc/about",
    "op_tos_uri": "https://localhost:8020/oidc/about",
    "registration_endpoint": "https://localhost:8020/oidc/register",
    "revocation_endpoint": "https://localhost:8020/oidc/revoke",
    "service_documentation": "https://localhost:8020/oidc/about",
    "token_endpoint": "https://localhost:8020/oidc/token",
    "userinfo_endpoint": "https://localhost:8020/oidc/userinfo",
    "device_authorization_endpoint": "https://localhost:8020/oidc/authorize/device",
    "token_endpoint_auth_methods_supported": [
        "client_secret_post",
        "client_secret_basic",
        "client_secret_jwt",
        "private_key_jwt",
        "none",
    ],
}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import collections.abc
import enum
import typing as ty

from keystoneauth1 import _utils as utils
from keystoneauth1 import access
from keystoneauth1 import exceptions
from keystoneauth1.identity import base
from keystoneauth1 import session as ks_session

_logger = utils.get_logger(__name__)


class Auth(base.BaseIdentityPlugin, metaclass=abc.ABCMeta):
    """Identity V2 Authentication Plugin.

    :param string auth_url: Identity service endpoint for authorization.
    :param string trust_id: Trust ID for trust scoping.
    :param string tenant_id: Tenant ID for project scoping.
    :param string tenant_name: Tenant name for project scoping.
    :param bool reauthenticate: Allow fetching a new token if the current one
                                is going to expire. (optional) default True
    """

    auth_url: str

    def __init__(
        self,
        auth_url: str,
        *,
        trust_id: ty.Optional[str] = None,
        tenant_id: ty.Optional[str] = None,
        tenant_name: ty.Optional[str] = None,
        reauthenticate: bool = True,
    ):
        super().__init__(auth_url=auth_url, reauthenticate=reauthenticate)

        self.trust_id = trust_id
        self.tenant_id = tenant_id
        self.tenant_name = tenant_name

    def get_auth_ref(self, session: ks_session.Session) -> access.AccessInfoV2:
        headers = {'Accept': 'application/json'}
        url = self.auth_url.rstrip('/') + '/tokens'
        params = {'auth': self.get_auth_data(headers)}

        if self.tenant_id:
            params['auth']['tenantId'] = self.tenant_id
        elif self.tenant_name:
            params['auth']['tenantName'] = self.tenant_name
        if self.trust_id:
            params['auth']['trust_id'] = self.trust_id

        _logger.debug('Making authentication request to %s', url)
        resp = session.post(
            url, json=params, headers=headers, authenticated=False, log=False
        )

        try:
            resp_data = resp.json()
        except ValueError:
            raise exceptions.InvalidResponse(response=resp)

        if 'access' not in resp_data:
            raise exceptions.InvalidResponse(response=resp)

        return access.AccessInfoV2(resp_data)

    @abc.abstractmethod
    def get_auth_data(
        self,
        headers: ty.Optional[collections.abc.MutableMapping[str, str]] = None,
    ) -> dict[str, object]:
        """Return the authentication section of an auth plugin.

        :param dict headers: The headers that will be sent with the auth
                             request if a plugin needs to add to them.
        :return: A dict of authentication data for the auth type.
        :rtype: dict
        """

    @property
    def has_scope_parameters(self) -> bool:
        """Return true if parameters can be used to create a scoped token."""
        return bool(self.tenant_id or self.tenant_name or self.trust_id)


# https://peps.python.org/pep-0484/#support-for-singleton-types-in-unions
class Unset(enum.Enum):
    token = 0


_unset = Unset.token


class Password(Auth):
    """A plugin for authenticating with a username and password.

    A username or user_id must be provided.

    :param string auth_url: Identity service endpoint for authorization.
    :param string username: Username for authentication.
    :param string password: Password for authentication.
    :param string user_id: User ID for authentication.
    :param string trust_id: Trust ID for trust scoping.
    :param string tenant_id: Tenant ID for tenant scoping.
    :param string tenant_name: Tenant name for tenant scoping.
    :param bool reauthenticate: Allow fetching a new token if the current one
                                is going to expire. (optional) default True

    :raises TypeError: if a user_id or username is not provided.
    """

    # FIXME(stephenfin): The use of _unset is a hack to work around
    # misconfiguration issues with random services (bug #1361444). It needs to
    # go away asap. See change Id61cfd1423afa8f9dd964fda278f4fab40887512 for
    # more info.
    def __init__(
        self,
        auth_url: str,
        username: ty.Union[str, None, Unset] = _unset,
        password: ty.Optional[str] = None,
        user_id: ty.Union[str, None, Unset] = _unset,
        *,
        trust_id: ty.Optional[str] = None,
        tenant_id: ty.Optional[str] = None,
        tenant_name: ty.Optional[str] = None,
        reauthenticate: bool = True,
    ):
        super().__init__(
            auth_url,
            trust_id=trust_id,
            tenant_id=tenant_id,
            tenant_name=tenant_name,
            reauthenticate=reauthenticate,
        )

        if username is _unset and user_id is _unset:
            msg = 'You need to specify either a username or user_id'
            raise TypeError(msg)

        if username is _unset:
            self.username = None
        else:
            self.username = username

        if user_id is _unset:
            self.user_id = None
        else:
            self.user_id = user_id

        self.password = password

    def get_auth_data(
        self,
        headers: ty.Optional[collections.abc.MutableMapping[str, str]] = None,
    ) -> dict[str, object]:
        auth = {'password': self.password}

        if self.username:
            auth['username'] = self.username
        elif self.user_id:
            auth['userId'] = self.user_id

        return {'passwordCredentials': auth}

    def get_cache_id_elements(self) -> dict[str, ty.Optional[str]]:
        return {
            'username': self.username,
            'user_id': self.user_id,
            'password': self.password,
            'auth_url': self.auth_url,
            'tenant_id': self.tenant_id,
            'tenant_name': self.tenant_name,
            'trust_id': self.trust_id,
        }


class Token(Auth):
    """A plugin for authenticating with an existing token.

    :param string auth_url: Identity service endpoint for authorization.
    :param string token: Existing token for authentication.
    :param string tenant_id: Tenant ID for tenant scoping.
    :param string tenant_name: Tenant name for tenant scoping.
    :param string trust_id: Trust ID for trust scoping.
    :param bool reauthenticate: Allow fetching a new token if the current one
                                is going to expire. (optional) default True
    """

    def __init__(
        self,
        auth_url: str,
        token: str,
        *,
        trust_id: ty.Optional[str] = None,
        tenant_id: ty.Optional[str] = None,
        tenant_name: ty.Optional[str] = None,
        reauthenticate: bool = True,
    ):
        super().__init__(
            auth_url,
            trust_id=trust_id,
            tenant_id=tenant_id,
            tenant_name=tenant_name,
            reauthenticate=reauthenticate,
        )
        self.token = token

    def get_auth_data(
        self,
        headers: ty.Optional[collections.abc.MutableMapping[str, str]] = None,
    ) -> dict[str, object]:
        if headers is not None:
            headers['X-Auth-Token'] = self.token
        return {'token': {'id': self.token}}

    def get_cache_id_elements(self) -> dict[str, ty.Optional[str]]:
        return {
            'token': self.token,
            'auth_url': self.auth_url,
            'tenant_id': self.tenant_id,
            'tenant_name': self.tenant_name,
            'trust_id': self.trust_id,
        }



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty

import os_service_types

from keystoneauth1.exceptions import base

_SERVICE_TYPES = os_service_types.ServiceTypes()


__all__ = (
    'DiscoveryFailure',
    'ImpliedVersionMismatch',
    'ImpliedMinVersionMismatch',
    'ImpliedMaxVersionMismatch',
    'VersionNotAvailable',
)

_PARSED_VERSION_T = tuple[ty.Union[int, float], ...]


class DiscoveryFailure(base.ClientException):
    message = "Discovery of client versions failed."


class VersionNotAvailable(DiscoveryFailure):
    message = "Discovery failed. Requested version is not available."


class ImpliedVersionMismatch(ValueError):
    label = 'version'

    def __init__(
        self, service_type: str, implied: _PARSED_VERSION_T, given: str
    ):
        super().__init__(
            f"service_type {service_type} was given which implies major API "
            f"version {str(implied[0])} but {self.label} of {given} was also "
            f"given. Please update your code to use the official service_type "
            f"{_SERVICE_TYPES.get_service_type(service_type)}."
        )


class ImpliedMinVersionMismatch(ImpliedVersionMismatch):
    label = 'min_version'


class ImpliedMaxVersionMismatch(ImpliedVersionMismatch):
    label = 'max_version'



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from keystoneauth1.identity.generic import token
from keystoneauth1.identity import v2
from keystoneauth1.identity import v3
from keystoneauth1.identity.v3 import token as v3_token
from keystoneauth1.tests.unit.identity import utils


class TokenTests(utils.GenericPluginTestCase):
    PLUGIN_CLASS = token.Token
    V2_PLUGIN_CLASS = v2.Token
    V3_PLUGIN_CLASS = v3.Token

    def new_plugin(self, **kwargs):
        kwargs.setdefault('token', uuid.uuid4().hex)
        return super().new_plugin(**kwargs)

    def test_symbols(self):
        self.assertIs(v3.Token, v3_token.Token)
        self.assertIs(v3.TokenMethod, v3_token.TokenMethod)

    def test_token_cache_id(self):
        the_token = uuid.uuid4().hex
        project_name = uuid.uuid4().hex
        default_domain_id = uuid.uuid4().hex

        a = token.Token(
            self.TEST_URL,
            token=the_token,
            project_name=project_name,
            default_domain_id=default_domain_id,
        )

        b = token.Token(
            self.TEST_URL,
            token=the_token,
            project_name=project_name,
            default_domain_id=default_domain_id,
        )

        a_id = a.get_cache_id()
        b_id = b.get_cache_id()

        self.assertEqual(a_id, b_id)

        c = token.Token(
            self.TEST_URL,
            token=the_token,
            project_name=uuid.uuid4().hex,  # different
            default_domain_id=default_domain_id,
        )

        c_id = c.get_cache_id()

        self.assertNotEqual(a_id, c_id)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty

if ty.TYPE_CHECKING:
    from oslo_config import cfg as _cfg

cfg = None
_NOT_FOUND = object()


def get_oslo_config() -> '_cfg':
    """Runtime load the oslo.config object.

    In performance optimization of openstackclient it was determined that even
    optimistically loading oslo.config if available had a performance cost.
    Given that we used to only raise the ImportError when the function was
    called also attempt to do the import to do everything at runtime.
    """
    global cfg

    # First Call
    if not cfg:
        try:
            from oslo_config import cfg
        except ImportError:
            cfg = _NOT_FOUND

    if cfg is _NOT_FOUND:
        raise ImportError(
            "oslo.config is not an automatic dependency of "
            "keystoneauth. If you wish to use oslo.config "
            "you need to import it into your application's "
            "requirements file. "
        )

    return cfg



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from testtools import matchers

from keystoneauth1.loading._plugins import admin_token as loader
from keystoneauth1 import session
from keystoneauth1.tests.unit import utils
from keystoneauth1 import token_endpoint


class TokenEndpointTest(utils.TestCase):
    TEST_TOKEN = 'aToken'
    TEST_URL = 'http://server/prefix'

    def test_basic_case(self):
        self.requests_mock.get(self.TEST_URL, text='body')

        a = token_endpoint.Token(self.TEST_URL, self.TEST_TOKEN)
        s = session.Session(auth=a)

        data = s.get(self.TEST_URL, authenticated=True)

        self.assertEqual(data.text, 'body')
        self.assertRequestHeaderEqual('X-Auth-Token', self.TEST_TOKEN)

    def test_basic_endpoint_case(self):
        self.stub_url('GET', ['p'], text='body')
        a = token_endpoint.Token(self.TEST_URL, self.TEST_TOKEN)
        s = session.Session(auth=a)

        data = s.get(
            '/p', authenticated=True, endpoint_filter={'service': 'identity'}
        )

        self.assertEqual(self.TEST_URL, a.get_endpoint(s))
        self.assertEqual('body', data.text)
        self.assertRequestHeaderEqual('X-Auth-Token', self.TEST_TOKEN)

    def test_token_endpoint_user_id(self):
        a = token_endpoint.Token(self.TEST_URL, self.TEST_TOKEN)
        s = session.Session()

        # we can't know this information about this sort of plugin
        self.assertIsNone(a.get_user_id(s))
        self.assertIsNone(a.get_project_id(s))


class AdminTokenTest(utils.TestCase):
    def test_token_endpoint_options(self):
        opt_names = [opt.name for opt in loader.AdminToken().get_options()]

        self.assertThat(opt_names, matchers.HasLength(2))

        self.assertIn('token', opt_names)
        self.assertIn('endpoint', opt_names)

    def test_token_endpoint_deprecated_options(self):
        endpoint_opt = [
            opt
            for opt in loader.AdminToken().get_options()
            if opt.name == 'endpoint'
        ][0]

        opt_names = [opt.name for opt in endpoint_opt.deprecated]

        self.assertEqual(['url'], opt_names)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from keystoneauth1 import access
from keystoneauth1 import exceptions
from keystoneauth1 import fixture
from keystoneauth1.tests.unit import utils


class ServiceCatalogTest(utils.TestCase):
    def setUp(self):
        super().setUp()

        self.AUTH_RESPONSE_BODY = fixture.V2Token(
            token_id='ab48a9efdfedb23ty3494',
            expires='2010-11-01T03:32:15-05:00',
            tenant_id='345',
            tenant_name='My Project',
            user_id='123',
            user_name='jqsmith',
            audit_chain_id=uuid.uuid4().hex,
        )

        self.AUTH_RESPONSE_BODY.add_role(id='234', name='compute:admin')
        role = self.AUTH_RESPONSE_BODY.add_role(
            id='235', name='object-store:admin'
        )
        role['tenantId'] = '1'

        s = self.AUTH_RESPONSE_BODY.add_service('compute', 'Cloud Servers')
        endpoint = s.add_endpoint(
            public='https://compute.north.host/v1/1234',
            internal='https://compute.north.host/v1/1234',
            region='North',
        )
        endpoint['tenantId'] = '1'
        endpoint['versionId'] = '1.0'
        endpoint['versionInfo'] = 'https://compute.north.host/v1.0/'
        endpoint['versionList'] = 'https://compute.north.host/'

        endpoint = s.add_endpoint(
            public='https://compute.north.host/v1.1/3456',
            internal='https://compute.north.host/v1.1/3456',
            region='North',
        )
        endpoint['tenantId'] = '2'
        endpoint['versionId'] = '1.1'
        endpoint['versionInfo'] = 'https://compute.north.host/v1.1/'
        endpoint['versionList'] = 'https://compute.north.host/'

        s = self.AUTH_RESPONSE_BODY.add_service('object-store', 'Cloud Files')
        endpoint = s.add_endpoint(
            public='https://swift.north.host/v1/blah',
            internal='https://swift.north.host/v1/blah',
            region='South',
        )
        endpoint['tenantId'] = '11'
        endpoint['versionId'] = '1.0'
        endpoint['versionInfo'] = 'uri'
        endpoint['versionList'] = 'uri'

        endpoint = s.add_endpoint(
            public='https://swift.north.host/v1.1/blah',
            internal='https://compute.north.host/v1.1/blah',
            region='South',
        )
        endpoint['tenantId'] = '2'
        endpoint['versionId'] = '1.1'
        endpoint['versionInfo'] = 'https://swift.north.host/v1.1/'
        endpoint['versionList'] = 'https://swift.north.host/'

        s = self.AUTH_RESPONSE_BODY.add_service('image', 'Image Servers')
        s.add_endpoint(
            public='https://image.north.host/v1/',
            internal='https://image-internal.north.host/v1/',
            region='North',
        )
        s.add_endpoint(
            public='https://image.south.host/v1/',
            internal='https://image-internal.south.host/v1/',
            region='South',
        )

    def test_building_a_service_catalog(self):
        auth_ref = access.create(body=self.AUTH_RESPONSE_BODY)
        sc = auth_ref.service_catalog

        self.assertEqual(
            sc.url_for(service_type='compute'),
            "https://compute.north.host/v1/1234",
        )
        self.assertRaises(
            exceptions.EndpointNotFound,
            sc.url_for,
            region_name="South",
            service_type='compute',
        )

    def test_service_catalog_endpoints(self):
        auth_ref = access.create(body=self.AUTH_RESPONSE_BODY)
        sc = auth_ref.service_catalog
        public_ep = sc.get_endpoints(
            service_type='compute', interface='publicURL'
        )
        self.assertEqual(public_ep['compute'][1]['tenantId'], '2')
        self.assertEqual(public_ep['compute'][1]['versionId'], '1.1')
        self.assertEqual(
            public_ep['compute'][1]['internalURL'],
            "https://compute.north.host/v1.1/3456",
        )

    def test_service_catalog_empty(self):
        self.AUTH_RESPONSE_BODY['access']['serviceCatalog'] = []
        auth_ref = access.create(body=self.AUTH_RESPONSE_BODY)
        self.assertRaises(
            exceptions.EmptyCatalog,
            auth_ref.service_catalog.url_for,
            service_type='image',
            interface='internalURL',
        )

    def test_service_catalog_get_endpoints_region_names(self):
        auth_ref = access.create(body=self.AUTH_RESPONSE_BODY)
        sc = auth_ref.service_catalog

        endpoints = sc.get_endpoints(service_type='image', region_name='North')
        self.assertEqual(len(endpoints), 1)
        self.assertEqual(
            endpoints['image'][0]['publicURL'], 'https://image.north.host/v1/'
        )

        endpoints = sc.get_endpoints(service_type='image', region_name='South')
        self.assertEqual(len(endpoints), 1)
        self.assertEqual(
            endpoints['image'][0]['publicURL'], 'https://image.south.host/v1/'
        )

        endpoints = sc.get_endpoints(service_type='compute')
        self.assertEqual(len(endpoints['compute']), 2)

        endpoints = sc.get_endpoints(
            service_type='compute', region_name='North'
        )
        self.assertEqual(len(endpoints['compute']), 2)

        endpoints = sc.get_endpoints(
            service_type='compute', region_name='West'
        )
        self.assertEqual(len(endpoints['compute']), 0)

    def test_service_catalog_url_for_region_names(self):
        auth_ref = access.create(body=self.AUTH_RESPONSE_BODY)
        sc = auth_ref.service_catalog

        url = sc.url_for(service_type='image', region_name='North')
        self.assertEqual(url, 'https://image.north.host/v1/')

        url = sc.url_for(service_type='image', region_name='South')
        self.assertEqual(url, 'https://image.south.host/v1/')

        self.assertRaises(
            exceptions.EndpointNotFound,
            sc.url_for,
            service_type='image',
            region_name='West',
        )

    def test_servcie_catalog_get_url_region_names(self):
        auth_ref = access.create(body=self.AUTH_RESPONSE_BODY)
        sc = auth_ref.service_catalog

        urls = sc.get_urls(service_type='image')
        self.assertEqual(len(urls), 2)

        urls = sc.get_urls(service_type='image', region_name='North')
        self.assertEqual(len(urls), 1)
        self.assertEqual(urls[0], 'https://image.north.host/v1/')

        urls = sc.get_urls(service_type='image', region_name='South')
        self.assertEqual(len(urls), 1)
        self.assertEqual(urls[0], 'https://image.south.host/v1/')

        urls = sc.get_urls(service_type='image', region_name='West')
        self.assertEqual(len(urls), 0)

    def test_service_catalog_service_name(self):
        auth_ref = access.create(body=self.AUTH_RESPONSE_BODY)
        sc = auth_ref.service_catalog

        url = sc.url_for(
            service_name='Image Servers',
            interface='public',
            service_type='image',
            region_name='North',
        )
        self.assertEqual('https://image.north.host/v1/', url)

        self.assertRaises(
            exceptions.EndpointNotFound,
            sc.url_for,
            service_name='Image Servers',
            service_type='compute',
        )

        urls = sc.get_urls(
            service_type='image',
            service_name='Image Servers',
            interface='public',
        )

        self.assertIn('https://image.north.host/v1/', urls)
        self.assertIn('https://image.south.host/v1/', urls)

        urls = sc.get_urls(
            service_type='image', service_name='Servers', interface='public'
        )

        self.assertEqual(0, len(urls))

    def test_service_catalog_multiple_service_types(self):
        token = fixture.V2Token()
        token.set_scope()

        for i in range(3):
            s = token.add_service('compute')
            s.add_endpoint(
                public=f'public-{i}',
                admin=f'admin-{i}',
                internal=f'internal-{i}',
                region=f'region-{i}',
            )

        auth_ref = access.create(body=token)

        urls = auth_ref.service_catalog.get_urls(
            service_type='compute', interface='publicURL'
        )

        self.assertEqual({'public-0', 'public-1', 'public-2'}, set(urls))

        urls = auth_ref.service_catalog.get_urls(
            service_type='compute',
            interface='publicURL',
            region_name='region-1',
        )

        self.assertEqual(('public-1',), urls)

    def test_service_catalog_endpoint_id(self):
        token = fixture.V2Token()
        token.set_scope()
        endpoint_id = uuid.uuid4().hex
        public_url = uuid.uuid4().hex

        s = token.add_service('compute')
        s.add_endpoint(public=public_url, id=endpoint_id)
        s.add_endpoint(public=uuid.uuid4().hex)

        auth_ref = access.create(body=token)

        # initially assert that we get back all our urls for a simple filter
        urls = auth_ref.service_catalog.get_urls(interface='public')
        self.assertEqual(2, len(urls))

        urls = auth_ref.service_catalog.get_urls(
            endpoint_id=endpoint_id, interface='public'
        )

        self.assertEqual((public_url,), urls)

        # with bad endpoint_id nothing should be found
        urls = auth_ref.service_catalog.get_urls(
            endpoint_id=uuid.uuid4().hex, interface='public'
        )

        self.assertEqual(0, len(urls))

        # we ignore a service_id because v2 doesn't know what it is
        urls = auth_ref.service_catalog.get_urls(
            endpoint_id=endpoint_id,
            service_id=uuid.uuid4().hex,
            interface='public',
        )

        self.assertEqual((public_url,), urls)

    def test_service_catalog_without_service_type(self):
        token = fixture.V2Token()
        token.set_scope()

        public_urls = []

        for i in range(0, 3):
            public_url = uuid.uuid4().hex
            public_urls.append(public_url)

            s = token.add_service(uuid.uuid4().hex)
            s.add_endpoint(public=public_url)

        auth_ref = access.create(body=token)
        urls = auth_ref.service_catalog.get_urls(
            service_type=None, interface='public'
        )

        self.assertEqual(3, len(urls))

        for p in public_urls:
            self.assertIn(p, urls)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import fixtures


class HackingCode(fixtures.Fixture):
    """A fixture to house the various code examples.

    Examples contains various keystoneauth hacking style checks.
    """

    oslo_namespace_imports = {
        'code': """
            import oslo.utils
            import oslo_utils
            import oslo.utils.encodeutils
            import oslo_utils.encodeutils
            from oslo import utils
            from oslo.utils import encodeutils
            from oslo_utils import encodeutils

            import oslo.serialization
            import oslo_serialization
            import oslo.serialization.jsonutils
            import oslo_serialization.jsonutils
            from oslo import serialization
            from oslo.serialization import jsonutils
            from oslo_serialization import jsonutils

            import oslo.config
            import oslo_config
            import oslo.config.cfg
            import oslo_config.cfg
            from oslo import config
            from oslo.config import cfg
            from oslo_config import cfg

            import oslo.i18n
            import oslo_i18n
            import oslo.i18n.log
            import oslo_i18n.log
            from oslo import i18n
            from oslo.i18n import log
            from oslo_i18n import log
        """,
        'expected_errors': [
            (1, 0, 'K333'),
            (3, 0, 'K333'),
            (5, 0, 'K333'),
            (6, 0, 'K333'),
            (9, 0, 'K333'),
            (11, 0, 'K333'),
            (13, 0, 'K333'),
            (14, 0, 'K333'),
            (17, 0, 'K333'),
            (19, 0, 'K333'),
            (21, 0, 'K333'),
            (22, 0, 'K333'),
            (25, 0, 'K333'),
            (27, 0, 'K333'),
            (29, 0, 'K333'),
            (30, 0, 'K333'),
        ],
    }



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from keystoneauth1 import fixture
from keystoneauth1 import identity
from keystoneauth1 import service_token
from keystoneauth1 import session
from keystoneauth1.tests.unit import utils


class ServiceTokenTests(utils.TestCase):
    TEST_URL = 'http://test.example.com/path/'
    USER_URL = 'http://user-keystone.example.com/v3'
    SERVICE_URL = 'http://service-keystone.example.com/v3'

    def setUp(self):
        super().setUp()

        self.user_token_id = uuid.uuid4().hex
        self.user_token = fixture.V3Token()
        self.user_token.set_project_scope()
        self.user_auth = identity.V3Password(
            auth_url=self.USER_URL,
            user_id=uuid.uuid4().hex,
            password=uuid.uuid4().hex,
            project_id=uuid.uuid4().hex,
        )

        self.service_token_id = uuid.uuid4().hex
        self.service_token = fixture.V3Token()
        self.service_token.set_project_scope()
        self.service_auth = identity.V3Password(
            auth_url=self.SERVICE_URL,
            user_id=uuid.uuid4().hex,
            password=uuid.uuid4().hex,
            project_id=uuid.uuid4().hex,
        )

        for t in (self.user_token, self.service_token):
            s = t.add_service('identity')
            s.add_standard_endpoints(
                public='http://keystone.example.com',
                admin='http://keystone.example.com',
                internal='http://keystone.example.com',
            )

        self.test_data = {'data': uuid.uuid4().hex}

        self.user_mock = self.requests_mock.post(
            self.USER_URL + '/auth/tokens',
            json=self.user_token,
            headers={'X-Subject-Token': self.user_token_id},
        )

        self.service_mock = self.requests_mock.post(
            self.SERVICE_URL + '/auth/tokens',
            json=self.service_token,
            headers={'X-Subject-Token': self.service_token_id},
        )

        self.requests_mock.get(self.TEST_URL, json=self.test_data)

        self.combined_auth = service_token.ServiceTokenAuthWrapper(
            self.user_auth, self.service_auth
        )

        self.session = session.Session(auth=self.combined_auth)

    def test_setting_service_token(self):
        self.session.get(self.TEST_URL)

        headers = self.requests_mock.last_request.headers

        self.assertEqual(self.user_token_id, headers['X-Auth-Token'])
        self.assertEqual(self.service_token_id, headers['X-Service-Token'])

        self.assertTrue(self.user_mock.called_once)
        self.assertTrue(self.service_mock.called_once)

    def test_invalidation(self):
        text = uuid.uuid4().hex
        test_url = 'http://test.example.com/abc'

        response_list = [{'status_code': 401}, {'text': text}]
        mock = self.requests_mock.get(test_url, response_list=response_list)

        resp = self.session.get(test_url)
        self.assertEqual(text, resp.text)

        self.assertEqual(2, mock.call_count)
        self.assertEqual(2, self.user_mock.call_count)
        self.assertEqual(2, self.service_mock.call_count)

    def test_pass_throughs(self):
        self.assertEqual(
            self.user_auth.get_token(self.session),
            self.combined_auth.get_token(self.session),
        )

        self.assertEqual(
            self.user_auth.get_endpoint(self.session, service_type='identity'),
            self.combined_auth.get_endpoint(
                self.session, service_type='identity'
            ),
        )

        self.assertEqual(
            self.user_auth.get_user_id(self.session),
            self.combined_auth.get_user_id(self.session),
        )

        self.assertEqual(
            self.user_auth.get_project_id(self.session),
            self.combined_auth.get_project_id(self.session),
        )

        self.assertEqual(
            self.user_auth.get_sp_auth_url(self.session, 'a'),
            self.combined_auth.get_sp_auth_url(self.session, 'a'),
        )

        self.assertEqual(
            self.user_auth.get_sp_url(self.session, 'a'),
            self.combined_auth.get_sp_url(self.session, 'a'),
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import typing as ty
import urllib
import uuid

try:
    # explicitly re-export symbol
    # https://mypy.readthedocs.io/en/stable/command_line.html#cmdoption-mypy-no-implicit-reexport
    from lxml import etree as etree
except ImportError:
    etree = None  # type: ignore[assignment]

from keystoneauth1 import access
from keystoneauth1 import exceptions
from keystoneauth1.extras._saml2.v3 import base
from keystoneauth1 import session as ks_session


class Password(base.BaseSAMLPlugin):
    """Authentication plugin for Microsoft ADFS2.0 IdPs."""

    DEFAULT_ADFS_TOKEN_EXPIRATION = 120

    HEADER_SOAP = {"Content-Type": "application/soap+xml; charset=utf-8"}
    HEADER_X_FORM = {"Content-Type": "application/x-www-form-urlencoded"}

    NAMESPACES = {
        's': 'http://www.w3.org/2003/05/soap-envelope',
        'a': 'http://www.w3.org/2005/08/addressing',
        'u': (
            'http://docs.oasis-open.org/wss/2004/01/oasis-200401-'
            'wss-wssecurity-utility-1.0.xsd'
        ),
    }

    ADFS_TOKEN_NAMESPACES = {
        's': 'http://www.w3.org/2003/05/soap-envelope',
        't': 'http://docs.oasis-open.org/ws-sx/ws-trust/200512',
    }
    ADFS_ASSERTION_XPATH = (
        '/s:Envelope/s:Body'
        '/t:RequestSecurityTokenResponseCollection'
        '/t:RequestSecurityTokenResponse'
    )

    def __init__(
        self,
        auth_url: str,
        identity_provider: str,
        identity_provider_url: str,
        service_provider_endpoint: str,
        username: str,
        password: str,
        protocol: str,
        service_provider_entity_id: ty.Optional[str] = None,
        *,
        trust_id: ty.Optional[str] = None,
        system_scope: ty.Optional[str] = None,
        domain_id: ty.Optional[str] = None,
        domain_name: ty.Optional[str] = None,
        project_id: ty.Optional[str] = None,
        project_name: ty.Optional[str] = None,
        project_domain_id: ty.Optional[str] = None,
        project_domain_name: ty.Optional[str] = None,
        reauthenticate: bool = True,
        include_catalog: bool = True,
    ):
        """Constructor for ``ADFSPassword``.

        :param auth_url: URL of the Identity Service
        :type auth_url: string

        :param identity_provider: name of the Identity Provider the client
                                  will authenticate against. This parameter
                                  will be used to build a dynamic URL used to
                                  obtain unscoped OpenStack token.
        :type identity_provider: string

        :param identity_provider_url: An Identity Provider URL, where the SAML2
                                      authentication request will be sent.
        :type identity_provider_url: string

        :param service_provider_endpoint: Endpoint where an assertion is being
            sent, for instance: ``https://host.domain/Shibboleth.sso/ADFS``
        :type service_provider_endpoint: string
        :param service_provider_entity_id: Service Provider SAML Entity ID
        :type service_provider_entity_id: string

        :param username: User's login
        :type username: string

        :param password: User's password
        :type password: string

        """
        super().__init__(
            auth_url=auth_url,
            identity_provider=identity_provider,
            identity_provider_url=identity_provider_url,
            username=username,
            password=password,
            protocol=protocol,
            trust_id=trust_id,
            system_scope=system_scope,
            domain_id=domain_id,
            domain_name=domain_name,
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
            reauthenticate=reauthenticate,
            include_catalog=include_catalog,
        )

        self.service_provider_endpoint = service_provider_endpoint
        self.service_provider_entity_id = service_provider_entity_id

    def _cookies(self, session: ks_session.Session) -> bool:
        """Check if cookie jar is not empty.

        keystoneauth1.session.Session object doesn't have a cookies attribute.
        We should then try fetching cookies from the underlying
        requests.Session object. If that fails too, there is something wrong
        and let Python raise the AttributeError.

        :param session: A session object to extract cookies from
        :returns: True if cookie jar is nonempty, False otherwise
        :raises AttributeError: in case cookies are not find anywhere
        """
        return bool(session.session.cookies)

    def _token_dates(
        self, fmt: str = '%Y-%m-%dT%H:%M:%S.%fZ'
    ) -> tuple[str, str]:
        """Calculate created and expires datetime objects.

        The method is going to be used for building ADFS Request Security
        Token message. Time interval between ``created`` and ``expires``
        dates is now static and equals to 120 seconds. ADFS security tokens
        should not be live too long, as currently ``keystoneauth1``
        doesn't have mechanisms for reusing such tokens (every time ADFS authn
        method is called, keystoneauth1 will login with the ADFS instance).

        :param fmt: Datetime format for specifying string format of a date.
                    It should not be changed if the method is going to be used
                    for building the ADFS security token request.
        :type fmt: string

        """
        date_created = datetime.datetime.now(datetime.timezone.utc).replace(
            tzinfo=None
        )
        date_expires = date_created + datetime.timedelta(
            seconds=self.DEFAULT_ADFS_TOKEN_EXPIRATION
        )
        return date_created.strftime(fmt), date_expires.strftime(fmt)

    def _prepare_adfs_request(self) -> None:
        """Build the ADFS Request Security Token SOAP message.

        Some values like username or password are inserted in the request.

        """
        WSS_SECURITY_NAMESPACE = {
            'o': (
                'http://docs.oasis-open.org/wss/2004/01/oasis-200401-'
                'wss-wssecurity-secext-1.0.xsd'
            )
        }

        TRUST_NAMESPACE = {
            'trust': 'http://docs.oasis-open.org/ws-sx/ws-trust/200512'
        }

        WSP_NAMESPACE = {'wsp': 'http://schemas.xmlsoap.org/ws/2004/09/policy'}

        WSA_NAMESPACE = {'wsa': 'http://www.w3.org/2005/08/addressing'}

        root = etree.Element(
            '{http://www.w3.org/2003/05/soap-envelope}Envelope',
            nsmap=self.NAMESPACES,
        )

        header = etree.SubElement(
            root, '{http://www.w3.org/2003/05/soap-envelope}Header'
        )
        action = etree.SubElement(
            header, "{http://www.w3.org/2005/08/addressing}Action"
        )
        action.set(
            "{http://www.w3.org/2003/05/soap-envelope}mustUnderstand", "1"
        )
        action.text = (
            'http://docs.oasis-open.org/ws-sx/ws-trust/200512/RST/Issue'
        )

        messageID = etree.SubElement(
            header, '{http://www.w3.org/2005/08/addressing}MessageID'
        )
        messageID.text = 'urn:uuid:' + uuid.uuid4().hex
        replyID = etree.SubElement(
            header, '{http://www.w3.org/2005/08/addressing}ReplyTo'
        )
        address = etree.SubElement(
            replyID, '{http://www.w3.org/2005/08/addressing}Address'
        )
        address.text = 'http://www.w3.org/2005/08/addressing/anonymous'

        to = etree.SubElement(
            header, '{http://www.w3.org/2005/08/addressing}To'
        )
        to.set("{http://www.w3.org/2003/05/soap-envelope}mustUnderstand", "1")

        security = etree.SubElement(
            header,
            '{http://docs.oasis-open.org/wss/2004/01/oasis-200401-'
            'wss-wssecurity-secext-1.0.xsd}Security',
            nsmap=WSS_SECURITY_NAMESPACE,
        )

        security.set(
            "{http://www.w3.org/2003/05/soap-envelope}mustUnderstand", "1"
        )

        timestamp = etree.SubElement(
            security,
            (
                '{http://docs.oasis-open.org/wss/2004/01/oasis-200401-'
                'wss-wssecurity-utility-1.0.xsd}Timestamp'
            ),
        )
        timestamp.set(
            (
                '{http://docs.oasis-open.org/wss/2004/01/oasis-200401-'
                'wss-wssecurity-utility-1.0.xsd}Id'
            ),
            '_0',
        )

        created = etree.SubElement(
            timestamp,
            (
                '{http://docs.oasis-open.org/wss/2004/01/oasis-200401-'
                'wss-wssecurity-utility-1.0.xsd}Created'
            ),
        )

        expires = etree.SubElement(
            timestamp,
            (
                '{http://docs.oasis-open.org/wss/2004/01/oasis-200401-'
                'wss-wssecurity-utility-1.0.xsd}Expires'
            ),
        )

        created.text, expires.text = self._token_dates()

        usernametoken = etree.SubElement(
            security,
            '{http://docs.oasis-open.org/wss/2004/01/oasis-200401-'
            'wss-wssecurity-secext-1.0.xsd}UsernameToken',
        )
        usernametoken.set(
            (
                '{http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-'
                'wssecurity-utility-1.0.xsd}u'
            ),
            f"uuid-{uuid.uuid4().hex}-1",
        )

        username = etree.SubElement(
            usernametoken,
            (
                '{http://docs.oasis-open.org/wss/2004/01/oasis-'
                '200401-wss-wssecurity-secext-1.0.xsd}Username'
            ),
        )
        password = etree.SubElement(
            usernametoken,
            (
                '{http://docs.oasis-open.org/wss/2004/01/oasis-'
                '200401-wss-wssecurity-secext-1.0.xsd}Password'
            ),
            Type=(
                'http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-'
                'username-token-profile-1.0#PasswordText'
            ),
        )

        body = etree.SubElement(
            root, "{http://www.w3.org/2003/05/soap-envelope}Body"
        )

        request_security_token = etree.SubElement(
            body,
            (
                '{http://docs.oasis-open.org/ws-sx/ws-trust/200512}'
                'RequestSecurityToken'
            ),
            nsmap=TRUST_NAMESPACE,
        )

        applies_to = etree.SubElement(
            request_security_token,
            '{http://schemas.xmlsoap.org/ws/2004/09/policy}AppliesTo',
            nsmap=WSP_NAMESPACE,
        )

        endpoint_reference = etree.SubElement(
            applies_to,
            '{http://www.w3.org/2005/08/addressing}EndpointReference',
            nsmap=WSA_NAMESPACE,
        )

        wsa_address = etree.SubElement(
            endpoint_reference, '{http://www.w3.org/2005/08/addressing}Address'
        )

        keytype = etree.SubElement(
            request_security_token,
            '{http://docs.oasis-open.org/ws-sx/ws-trust/200512}KeyType',
        )
        keytype.text = (
            'http://docs.oasis-open.org/ws-sx/ws-trust/200512/Bearer'
        )

        request_type = etree.SubElement(
            request_security_token,
            '{http://docs.oasis-open.org/ws-sx/ws-trust/200512}RequestType',
        )
        request_type.text = (
            'http://docs.oasis-open.org/ws-sx/ws-trust/200512/Issue'
        )
        token_type = etree.SubElement(
            request_security_token,
            '{http://docs.oasis-open.org/ws-sx/ws-trust/200512}TokenType',
        )
        token_type.text = 'urn:oasis:names:tc:SAML:1.0:assertion'

        # After constructing the request, let's plug in some values
        username.text = self.username
        password.text = self.password
        to.text = self.identity_provider_url
        wsa_address.text = (
            self.service_provider_entity_id or self.service_provider_endpoint
        )

        self.prepared_request = root

    def _get_adfs_security_token(self, session: ks_session.Session) -> None:
        """Send ADFS Security token to the ADFS server.

        Store the result in the instance attribute and raise an exception in
        case the response is not valid XML data.

        If a user cannot authenticate due to providing bad credentials, the
        ADFS2.0 server will return a HTTP 500 response and a XML Fault message.
        If ``exceptions.InternalServerError`` is caught, the method tries to
        parse the XML response.
        If parsing is unsuccessful, an ``exceptions.AuthorizationFailure`` is
        raised with a reason from the XML fault. Otherwise an original
        ``exceptions.InternalServerError`` is re-raised.

        :param session : a session object to send out HTTP requests.
        :type session: keystoneauth1.session.Session

        :raises keystoneauth1.exceptions.AuthorizationFailure: when HTTP
                 response from the ADFS server is not a valid XML ADFS security
                 token.
        :raises keystoneauth1.exceptions.InternalServerError: If response
                 status code is HTTP 500 and the response XML cannot be
                 recognized.

        """

        request_security_token = self.xml_to_str(self.prepared_request)
        try:
            response = session.post(
                url=self.identity_provider_url,
                headers=self.HEADER_SOAP,
                data=request_security_token,
                authenticated=False,
            )
        except exceptions.InternalServerError as e:
            if e.response is None:
                raise

            try:
                obj = self.str_to_xml(e.response.content).xpath(
                    '/s:Envelope/s:Body/s:Fault/s:Code/s:Subcode/s:Value',
                    namespaces=self.NAMESPACES,
                )
                obj = self._first(obj)
                reason = obj.text
            # NOTE(marek-denis): etree.Element.xpath() doesn't raise an
            # exception, it just returns an empty list. In that case, _first()
            # will raise IndexError and we should treat it as an indication XML
            # is not valid. exceptions.AuthorizationFailure can be raised from
            # str_to_xml(), however since server returned HTTP 500 we should
            # re-raise exceptions.InternalServerError.
            except (IndexError, exceptions.AuthorizationFailure):
                raise e
            raise exceptions.AuthorizationFailure(reason)
        msg = (
            'Error parsing XML returned from '
            'the ADFS Identity Provider, reason: %s'
        )
        self.adfs_token = self.str_to_xml(response.content, msg)

    def _prepare_sp_request(self) -> None:
        """Prepare ADFS Security Token to be sent to the Service Provider.

        The method works as follows:
        * Extract SAML2 assertion from the ADFS Security Token.
        * Replace namespaces
        * urlencode assertion
        * concatenate static string with the encoded assertion

        """
        assertion = self.adfs_token.xpath(
            self.ADFS_ASSERTION_XPATH, namespaces=self.ADFS_TOKEN_NAMESPACES
        )
        assertion = self._first(assertion)
        assertion = self.xml_to_str(assertion)
        # TODO(marek-denis): Ideally no string replacement should occur.
        # Unfortunately lxml doesn't allow for namespaces changing in-place and
        # probably the only solution good for now is to build the assertion
        # from scratch and reuse values from the adfs security token.
        assertion = assertion.replace(
            b'http://docs.oasis-open.org/ws-sx/ws-trust/200512',
            b'http://schemas.xmlsoap.org/ws/2005/02/trust',
        )

        encoded_assertion = urllib.parse.quote(assertion)
        self.encoded_assertion = 'wa=wsignin1.0&wresult=' + encoded_assertion

    def _send_assertion_to_service_provider(
        self, session: ks_session.Session
    ) -> None:
        """Send prepared assertion to a service provider.

        As the assertion doesn't contain a protected resource, the value from
        the ``location`` header is not valid and we should not let the Session
        object get redirected there. The aim of this call is to get a cookie in
        the response which is required for entering a protected endpoint.

        :param session : a session object to send out HTTP requests.
        :type session: keystoneauth1.session.Session

        :raises: Corresponding HTTP error exception

        """
        session.post(
            url=self.service_provider_endpoint,
            data=self.encoded_assertion,
            headers=self.HEADER_X_FORM,
            redirect=False,
            authenticated=False,
        )

    def _access_service_provider(self, session: ks_session.Session) -> None:
        """Access protected endpoint and fetch unscoped token.

        After federated authentication workflow a protected endpoint should be
        accessible with the session object. The access is granted basing on the
        cookies stored within the session object. If, for some reason no
        cookies are present (quantity test) it means something went wrong and
        user will not be able to fetch an unscoped token. In that case an
        ``exceptions.AuthorizationFailure` exception is raised and no HTTP call
        is even made.

        :param session : a session object to send out HTTP requests.
        :type session: keystoneauth1.session.Session

        :raises keystoneauth1.exceptions.AuthorizationFailure: in case session
        object has empty cookie jar.

        """
        if self._cookies(session) is False:
            raise exceptions.AuthorizationFailure(
                "Session object doesn't contain a cookie, therefore you are "
                "not allowed to enter the Identity Provider's protected area."
            )
        self.authenticated_response = session.get(
            self.federated_token_url, authenticated=False
        )

    def get_unscoped_auth_ref(
        self, session: ks_session.Session
    ) -> access.AccessInfoV3:
        """Retrieve unscoped token after authentcation with ADFS server.

        This is a multistep process:

        * Prepare ADFS Request Securty Token -
          build an etree.XML object filling certain attributes with proper user
          credentials, created/expires dates (ticket is be valid for 120
          seconds as currently we don't handle reusing ADFS issued security
          tokens).

        * Send ADFS Security token to the ADFS server. Step handled by

        * Receive and parse security token, extract actual SAML assertion and
          prepare a request addressed for the Service Provider endpoint.
          This also includes changing namespaces in the XML document. Step
          handled by ``ADFSPassword._prepare_sp_request()`` method.

        * Send prepared assertion to the Service Provider endpoint. Usually
          the server will respond with HTTP 301 code which should be ignored as
          the 'location' header doesn't contain protected area. The goal of
          this operation is fetching the session cookie which later allows for
          accessing protected URL endpoints. Step handed by
          ``ADFSPassword._send_assertion_to_service_provider()`` method.

        * Once the session cookie is issued, the protected endpoint can be
          accessed and an unscoped token can be retrieved. Step handled by
          ``ADFSPassword._access_service_provider()`` method.

        :param session: a session object to send out HTTP requests.
        :type session: keystoneauth1.session.Session

        :returns: AccessInfo
        :rtype: :py:class:`keystoneauth1.access.AccessInfo`

        """
        self._prepare_adfs_request()
        self._get_adfs_security_token(session)
        self._prepare_sp_request()
        self._send_assertion_to_service_provider(session)
        self._access_service_provider(session)

        access_info = access.create(resp=self.authenticated_response)
        assert isinstance(access_info, access.AccessInfoV3)  # nosec B101
        return access_info



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from keystoneauth1 import exceptions
from keystoneauth1.identity.v3 import tokenless_auth
from keystoneauth1 import session
from keystoneauth1.tests.unit import utils


class TokenlessAuthTest(utils.TestCase):
    TEST_URL = 'http://server/prefix'

    def create(
        self,
        auth_url,
        domain_id=None,
        domain_name=None,
        project_id=None,
        project_name=None,
        project_domain_id=None,
        project_domain_name=None,
    ):
        self.requests_mock.get(self.TEST_URL)
        auth = tokenless_auth.TokenlessAuth(
            auth_url=self.TEST_URL,
            domain_id=domain_id,
            domain_name=domain_name,
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
        )
        return auth, session.Session(auth=auth)

    def test_domain_id_scope_header_pass(self):
        domain_id = uuid.uuid4().hex
        auth, session = self.create(
            auth_url=self.TEST_URL, domain_id=domain_id
        )
        session.get(self.TEST_URL, authenticated=True)
        self.assertRequestHeaderEqual('X-Domain-Id', domain_id)

    def test_domain_name_scope_header_pass(self):
        domain_name = uuid.uuid4().hex
        auth, session = self.create(
            auth_url=self.TEST_URL, domain_name=domain_name
        )
        session.get(self.TEST_URL, authenticated=True)
        self.assertRequestHeaderEqual('X-Domain-Name', domain_name)

    def test_project_id_scope_header_pass(self):
        project_id = uuid.uuid4().hex
        auth, session = self.create(
            auth_url=self.TEST_URL, project_id=project_id
        )
        session.get(self.TEST_URL, authenticated=True)
        self.assertRequestHeaderEqual('X-Project-Id', project_id)

    def test_project_of_domain_id_scope_header_pass(self):
        project_name = uuid.uuid4().hex
        project_domain_id = uuid.uuid4().hex
        auth, session = self.create(
            auth_url=self.TEST_URL,
            project_name=project_name,
            project_domain_id=project_domain_id,
        )
        session.get(self.TEST_URL, authenticated=True)
        self.assertRequestHeaderEqual('X-Project-Name', project_name)
        self.assertRequestHeaderEqual('X-Project-Domain-Id', project_domain_id)

    def test_project_of_domain__name_scope_header_pass(self):
        project_name = uuid.uuid4().hex
        project_domain_name = uuid.uuid4().hex
        auth, session = self.create(
            auth_url=self.TEST_URL,
            project_name=project_name,
            project_domain_name=project_domain_name,
        )
        session.get(self.TEST_URL, authenticated=True)
        self.assertRequestHeaderEqual('X-Project-Name', project_name)
        self.assertRequestHeaderEqual(
            'X-Project-Domain-Name', project_domain_name
        )

    def test_no_scope_header_fail(self):
        auth, session = self.create(auth_url=self.TEST_URL)
        self.assertIsNone(auth.get_headers(session))
        msg = 'No valid authentication is available'
        self.assertRaisesRegex(
            exceptions.AuthorizationFailure,
            msg,
            session.get,
            self.TEST_URL,
            authenticated=True,
        )

    def test_project_name_scope_only_header_fail(self):
        project_name = uuid.uuid4().hex
        auth, session = self.create(
            auth_url=self.TEST_URL, project_name=project_name
        )
        self.assertIsNone(auth.get_headers(session))
        msg = 'No valid authentication is available'
        self.assertRaisesRegex(
            exceptions.AuthorizationFailure,
            msg,
            session.get,
            self.TEST_URL,
            authenticated=True,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import urllib
import uuid

from lxml import etree

from keystoneauth1 import exceptions
from keystoneauth1.extras import _saml2 as saml2
from keystoneauth1.tests.unit import client_fixtures
from keystoneauth1.tests.unit.extras.saml2 import fixtures as saml2_fixtures
from keystoneauth1.tests.unit.extras.saml2 import utils
from keystoneauth1.tests.unit import matchers


class AuthenticateviaADFSTests(utils.TestCase):
    GROUP = 'auth'

    NAMESPACES = {
        's': 'http://www.w3.org/2003/05/soap-envelope',
        'trust': 'http://docs.oasis-open.org/ws-sx/ws-trust/200512',
        'wsa': 'http://www.w3.org/2005/08/addressing',
        'wsp': 'http://schemas.xmlsoap.org/ws/2004/09/policy',
        'a': 'http://www.w3.org/2005/08/addressing',
        'o': (
            'http://docs.oasis-open.org/wss/2004/01/oasis'
            '-200401-wss-wssecurity-secext-1.0.xsd'
        ),
    }

    USER_XPATH = '/s:Envelope/s:Header/o:Security/o:UsernameToken/o:Username'
    PASSWORD_XPATH = (
        '/s:Envelope/s:Header/o:Security/o:UsernameToken/o:Password'
    )
    ADDRESS_XPATH = (
        '/s:Envelope/s:Body'
        '/trust:RequestSecurityToken'
        '/wsp:AppliesTo/wsa:EndpointReference'
        '/wsa:Address'
    )
    TO_XPATH = '/s:Envelope/s:Header/a:To'

    TEST_TOKEN = uuid.uuid4().hex

    PROTOCOL = 'saml2'

    @property
    def _uuid4(self):
        return '4b911420-4982-4009-8afc-5c596cd487f5'

    def setUp(self):
        super().setUp()

        self.IDENTITY_PROVIDER = 'adfs'
        self.IDENTITY_PROVIDER_URL = (
            'http://adfs.local/adfs/service/trust/13/usernamemixed'
        )
        self.FEDERATION_AUTH_URL = '{}/{}'.format(
            self.TEST_URL,
            'OS-FEDERATION/identity_providers/adfs/protocols/saml2/auth',
        )
        self.SP_ENDPOINT = 'https://openstack4.local/Shibboleth.sso/ADFS'
        self.SP_ENTITYID = 'https://openstack4.local'

        self.adfsplugin = saml2.V3ADFSPassword(
            self.TEST_URL,
            self.IDENTITY_PROVIDER,
            self.IDENTITY_PROVIDER_URL,
            self.SP_ENDPOINT,
            self.TEST_USER,
            self.TEST_TOKEN,
            self.PROTOCOL,
        )

        self.ADFS_SECURITY_TOKEN_RESPONSE = utils._load_xml(
            'ADFS_RequestSecurityTokenResponse.xml'
        )
        self.ADFS_FAULT = utils._load_xml('ADFS_fault.xml')

    def test_get_adfs_security_token(self):
        """Test ADFSPassword._get_adfs_security_token()."""
        self.requests_mock.post(
            self.IDENTITY_PROVIDER_URL,
            content=utils.make_oneline(self.ADFS_SECURITY_TOKEN_RESPONSE),
            status_code=200,
        )

        self.adfsplugin._prepare_adfs_request()
        self.adfsplugin._get_adfs_security_token(self.session)

        adfs_response = etree.tostring(self.adfsplugin.adfs_token)
        fixture_response = self.ADFS_SECURITY_TOKEN_RESPONSE

        self.assertThat(fixture_response, matchers.XMLEquals(adfs_response))

    def test_adfs_request_user(self):
        self.adfsplugin._prepare_adfs_request()
        user = self.adfsplugin.prepared_request.xpath(
            self.USER_XPATH, namespaces=self.NAMESPACES
        )[0]
        self.assertEqual(self.TEST_USER, user.text)

    def test_adfs_request_password(self):
        self.adfsplugin._prepare_adfs_request()
        password = self.adfsplugin.prepared_request.xpath(
            self.PASSWORD_XPATH, namespaces=self.NAMESPACES
        )[0]
        self.assertEqual(self.TEST_TOKEN, password.text)

    def test_adfs_request_to(self):
        self.adfsplugin._prepare_adfs_request()
        to = self.adfsplugin.prepared_request.xpath(
            self.TO_XPATH, namespaces=self.NAMESPACES
        )[0]
        self.assertEqual(self.IDENTITY_PROVIDER_URL, to.text)

    def test_prepare_adfs_request_address(self):
        self.adfsplugin._prepare_adfs_request()
        address = self.adfsplugin.prepared_request.xpath(
            self.ADDRESS_XPATH, namespaces=self.NAMESPACES
        )[0]
        self.assertEqual(self.SP_ENDPOINT, address.text)

    def test_prepare_adfs_request_custom_endpointreference(self):
        self.adfsplugin = saml2.V3ADFSPassword(
            self.TEST_URL,
            self.IDENTITY_PROVIDER,
            self.IDENTITY_PROVIDER_URL,
            self.SP_ENDPOINT,
            self.TEST_USER,
            self.TEST_TOKEN,
            self.PROTOCOL,
            self.SP_ENTITYID,
        )
        self.adfsplugin._prepare_adfs_request()
        address = self.adfsplugin.prepared_request.xpath(
            self.ADDRESS_XPATH, namespaces=self.NAMESPACES
        )[0]
        self.assertEqual(self.SP_ENTITYID, address.text)

    def test_prepare_sp_request(self):
        assertion = etree.XML(self.ADFS_SECURITY_TOKEN_RESPONSE)
        assertion = assertion.xpath(
            saml2.V3ADFSPassword.ADFS_ASSERTION_XPATH,
            namespaces=saml2.V3ADFSPassword.ADFS_TOKEN_NAMESPACES,
        )
        assertion = assertion[0]
        assertion = etree.tostring(assertion)

        assertion = assertion.replace(
            b'http://docs.oasis-open.org/ws-sx/ws-trust/200512',
            b'http://schemas.xmlsoap.org/ws/2005/02/trust',
        )
        assertion = urllib.parse.quote(assertion)
        assertion = 'wa=wsignin1.0&wresult=' + assertion

        self.adfsplugin.adfs_token = etree.XML(
            self.ADFS_SECURITY_TOKEN_RESPONSE
        )
        self.adfsplugin._prepare_sp_request()

        self.assertEqual(assertion, self.adfsplugin.encoded_assertion)

    def test_get_adfs_security_token_authn_fail(self):
        """Test proper parsing XML fault after bad authentication.

        An exceptions.AuthorizationFailure should be raised including
        error message from the XML message indicating where was the problem.
        """
        content = utils.make_oneline(self.ADFS_FAULT)
        self.requests_mock.register_uri(
            'POST',
            self.IDENTITY_PROVIDER_URL,
            content=content,
            status_code=500,
        )

        self.adfsplugin._prepare_adfs_request()
        self.assertRaises(
            exceptions.AuthorizationFailure,
            self.adfsplugin._get_adfs_security_token,
            self.session,
        )
        # TODO(marek-denis): Python3 tests complain about missing 'message'
        # attributes
        # self.assertEqual('a:FailedAuthentication', e.message)

    def test_get_adfs_security_token_bad_response(self):
        """Test proper handling HTTP 500 and mangled (non XML) response.

        This should never happen yet, keystoneauth1 should be prepared
        and correctly raise exceptions.InternalServerError once it cannot
        parse XML fault message
        """
        self.requests_mock.register_uri(
            'POST',
            self.IDENTITY_PROVIDER_URL,
            content=b'NOT XML',
            status_code=500,
        )
        self.adfsplugin._prepare_adfs_request()
        self.assertRaises(
            exceptions.InternalServerError,
            self.adfsplugin._get_adfs_security_token,
            self.session,
        )

    # TODO(marek-denis): Need to figure out how to properly send cookies
    # from the request_mock methods.
    def _send_assertion_to_service_provider(self):
        """Test whether SP issues a cookie."""
        cookie = uuid.uuid4().hex

        self.requests_mock.post(
            self.SP_ENDPOINT, headers={"set-cookie": cookie}, status_code=302
        )

        self.adfsplugin.adfs_token = self._build_adfs_request()
        self.adfsplugin._prepare_sp_request()
        self.adfsplugin._send_assertion_to_service_provider(self.session)

        self.assertEqual(1, len(self.session.session.cookies))

    def test_send_assertion_to_service_provider_bad_status(self):
        self.requests_mock.register_uri(
            'POST', self.SP_ENDPOINT, status_code=500
        )

        self.adfsplugin.adfs_token = etree.XML(
            self.ADFS_SECURITY_TOKEN_RESPONSE
        )
        self.adfsplugin._prepare_sp_request()

        self.assertRaises(
            exceptions.InternalServerError,
            self.adfsplugin._send_assertion_to_service_provider,
            self.session,
        )

    def test_access_sp_no_cookies_fail(self):
        # clean cookie jar
        self.session.session.cookies = []

        self.assertRaises(
            exceptions.AuthorizationFailure,
            self.adfsplugin._access_service_provider,
            self.session,
        )

    def test_check_valid_token_when_authenticated(self):
        self.requests_mock.register_uri(
            'GET',
            self.FEDERATION_AUTH_URL,
            json=saml2_fixtures.UNSCOPED_TOKEN,
            headers=client_fixtures.AUTH_RESPONSE_HEADERS,
        )

        self.session.session.cookies = [object()]
        self.adfsplugin._access_service_provider(self.session)
        response = self.adfsplugin.authenticated_response

        self.assertEqual(
            client_fixtures.AUTH_RESPONSE_HEADERS, response.headers
        )

        self.assertEqual(
            saml2_fixtures.UNSCOPED_TOKEN['token'], response.json()['token']
        )

    def test_end_to_end_workflow(self):
        self.requests_mock.register_uri(
            'POST',
            self.IDENTITY_PROVIDER_URL,
            content=self.ADFS_SECURITY_TOKEN_RESPONSE,
            status_code=200,
        )
        self.requests_mock.register_uri(
            'POST',
            self.SP_ENDPOINT,
            headers={"set-cookie": 'x'},
            status_code=302,
        )
        self.requests_mock.register_uri(
            'GET',
            self.FEDERATION_AUTH_URL,
            json=saml2_fixtures.UNSCOPED_TOKEN,
            headers=client_fixtures.AUTH_RESPONSE_HEADERS,
        )

        # NOTE(marek-denis): We need to mimic this until self.requests_mock can
        # issue cookies properly.
        self.session.session.cookies = [object()]
        token = self.adfsplugin.get_auth_ref(self.session)
        self.assertEqual(client_fixtures.AUTH_SUBJECT_TOKEN, token.auth_token)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing as ty

import typing_extensions as ty_ext

# Identity V2 auth fields


class TenantV2(ty.TypedDict):
    description: ty_ext.NotRequired[ty.Optional[str]]
    enabled: ty_ext.NotRequired[bool]
    id: str
    name: str


class TokenV2(ty.TypedDict):
    audit_ids: list[str]
    bind: ty_ext.NotRequired[dict[str, ty.Any]]
    expires: str
    id: str
    issued_at: ty_ext.NotRequired[str]
    tenant: ty_ext.NotRequired[TenantV2]


class EndpointV2(ty.TypedDict):
    adminURL: str
    region: str
    internalURL: str
    id: str
    publicURL: str


class CatalogServiceV2(ty.TypedDict):
    endpoints: list[EndpointV2]
    endpoints_links: list[ty.Any]
    type: str
    name: str


class RoleV2(ty.TypedDict):
    name: str


class UserV2(ty.TypedDict):
    id: str
    name: str
    role_links: list[ty.Any]
    roles: list[RoleV2]
    tenantId: ty_ext.NotRequired[str]
    tenantName: ty_ext.NotRequired[str]
    username: str


class MetadataV2(ty.TypedDict):
    is_admin: int
    roles: list[str]


class TrustV2(ty.TypedDict):
    id: str
    impersonation: bool
    trustee_user_id: str
    trustor_user_id: str


class AccessV2(ty.TypedDict):
    token: TokenV2
    serviceCatalog: ty_ext.NotRequired[list[CatalogServiceV2]]
    user: UserV2
    metadata: ty_ext.NotRequired[MetadataV2]
    trust: ty_ext.NotRequired[TrustV2]


class TokenResponseV2(ty.TypedDict):
    access: AccessV2


# Identity V3 auth fields


class EndpointV3(ty.TypedDict):
    id: str
    interface: str
    region: str
    region_id: str
    url: str


class ServiceV3(ty.TypedDict):
    endpoints: list[EndpointV3]
    id: str
    name: str
    type: str


class ProjectDomainV3(ty.TypedDict):
    id: str
    name: str


class ProjectV3(ty.TypedDict):
    domain: ProjectDomainV3
    id: str
    name: str


class DomainV3(ty.TypedDict):
    id: str
    name: str


class UserDomainV3(ty.TypedDict):
    id: str
    name: str


class FederationGroupV3(ty.TypedDict):
    id: str


class FederationProviderV3(ty.TypedDict):
    id: str


class FederationProtocolV3(ty.TypedDict):
    id: str


class FederationV3(ty.TypedDict):
    groups: list[FederationGroupV3]
    identity_provider: FederationProviderV3
    protocol: FederationProtocolV3


UserV3 = ty.TypedDict(
    'UserV3',
    {
        'domain': UserDomainV3,
        'id': str,
        'name': str,
        'password_expires_at': ty_ext.NotRequired[str],
        'OS-FEDERATION': ty_ext.NotRequired[FederationV3],
    },
)


class RoleV3(ty.TypedDict):
    id: str
    name: str


class ApplicationCredentialAccessRuleV3(ty.TypedDict):
    id: str


class ApplicationCredentialV3(ty.TypedDict):
    access_rules: ty_ext.NotRequired[list[ApplicationCredentialAccessRuleV3]]
    id: str
    name: str
    restricted: bool


class ServiceProviderV3(ty.TypedDict):
    auth_url: str
    id: str
    sp_url: str


class TrustorUser(ty.TypedDict):
    id: str


class TrusteeUser(ty.TypedDict):
    id: str


class TrustV3(ty.TypedDict):
    id: str
    impersonation: bool
    trustee_user: TrusteeUser
    trustor_user: TrustorUser


class OAuth1V3(ty.TypedDict):
    access_token_id: str
    consumer_id: str


OAuth2V3 = ty.TypedDict('OAuth2V3', {'x5t#S256': str})


class SystemV3(ty.TypedDict):
    all: bool


TokenV3 = ty.TypedDict(
    'TokenV3',
    {
        'application_credential': ty_ext.NotRequired[ApplicationCredentialV3],
        'audit_ids': list[str],
        'bind': ty_ext.NotRequired[dict[str, ty.Any]],
        'catalog': ty_ext.NotRequired[list[ServiceV3]],
        'domain': ty_ext.NotRequired[DomainV3],
        'expires_at': str,
        'is_admin_project': ty_ext.NotRequired[bool],
        'is_domain': ty_ext.NotRequired[bool],
        'issued_at': str,
        'methods': list[str],
        'oauth2_credential': ty_ext.NotRequired[OAuth2V3],
        'project': ProjectV3,
        'roles': list[RoleV3],
        'service_providers': ty_ext.NotRequired[list[ServiceProviderV3]],
        'system': ty_ext.NotRequired[SystemV3],
        'user': UserV3,
        'OS-OAUTH1': ty_ext.NotRequired[OAuth1V3],
        'OS-TRUST:trust': ty_ext.NotRequired[TrustV3],
    },
)


class TokenResponseV3(ty.TypedDict):
    token: TokenV3



# Copyright 2022 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import typing as ty

from keystoneauth1 import access
from keystoneauth1 import exceptions
from keystoneauth1.identity.v3 import base
from keystoneauth1 import session as ks_session


class OAuth2mTlsClientCredential(base.BaseAuth, metaclass=abc.ABCMeta):
    """A plugin for authenticating via an OAuth2.0 mTLS client credential.

    :param string auth_url: keystone authorization endpoint.
    :param string oauth2_endpoint: OAuth2.0 endpoint.
    :param string oauth2_client_id: OAuth2.0 client credential id.
    """

    def __init__(
        self,
        auth_url: str,
        oauth2_endpoint: str,
        oauth2_client_id: str,
        *,
        trust_id: ty.Optional[str] = None,
        system_scope: ty.Optional[str] = None,
        domain_id: ty.Optional[str] = None,
        domain_name: ty.Optional[str] = None,
        project_id: ty.Optional[str] = None,
        project_name: ty.Optional[str] = None,
        project_domain_id: ty.Optional[str] = None,
        project_domain_name: ty.Optional[str] = None,
        reauthenticate: bool = True,
        include_catalog: bool = True,
    ):
        super().__init__(
            auth_url=auth_url,
            trust_id=trust_id,
            system_scope=system_scope,
            domain_id=domain_id,
            domain_name=domain_name,
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
            reauthenticate=reauthenticate,
            include_catalog=include_catalog,
        )
        self.oauth2_endpoint = oauth2_endpoint
        self.oauth2_client_id = oauth2_client_id
        self.oauth2_access_token = None

    def get_auth_ref(self, session: ks_session.Session) -> access.AccessInfoV3:
        """Obtain a token from an OpenStack Identity Service.

        This method is overridden by the various token version plugins.

        This function should not be called independently and is expected to be
        invoked via the do_authenticate function.

        This function will be invoked if the AcessInfo object cached by the
        plugin is not valid. Thus plugins should always fetch a new AccessInfo
        when invoked. If you are looking to just retrieve the current auth
        data then you should use get_access.

        :param session: A session object that can be used for communication.
        :type session: keystoneauth1.session.Session

        :raises keystoneauth1.exceptions.response.InvalidResponse:
            The response returned wasn't appropriate.
        :raises keystoneauth1.exceptions.http.HttpError:
            An error from an invalid HTTP response.
        :raises keystoneauth1.exceptions.ClientException:
            An error from getting OAuth2.0 access token.

        :returns: Token access information.
        :rtype: :class:`keystoneauth1.access.AccessInfo`
        """
        # Get OAuth2.0 access token and add the field 'Authorization' when
        # using the HTTPS protocol.
        data = {
            'grant_type': 'client_credentials',
            'client_id': self.oauth2_client_id,
        }
        resp = session.post(
            url=self.oauth2_endpoint,
            authenticated=False,
            raise_exc=False,
            data=data,
        )
        if resp.status_code == 200:
            oauth2 = resp.json()
            self.oauth2_access_token = oauth2.get('access_token')
        else:
            error = resp.json()
            msg = error.get('error_description')
            raise exceptions.ClientException(msg)

        headers = {
            'Accept': 'application/json',
            'X-Auth-Token': self.oauth2_access_token,
            'X-Subject-Token': self.oauth2_access_token,
        }

        token_url = '{}/auth/tokens'.format(self.auth_url.rstrip('/'))
        if not self.auth_url.rstrip('/').endswith('v3'):
            token_url = '{}/v3/auth/tokens'.format(self.auth_url.rstrip('/'))
        resp = session.get(
            url=token_url, authenticated=False, headers=headers, log=False
        )
        try:
            resp_data = resp.json()
        except ValueError:
            raise exceptions.InvalidResponse(response=resp)
        if 'token' not in resp_data:
            raise exceptions.InvalidResponse(response=resp)

        return access.AccessInfoV3(
            auth_token=self.oauth2_access_token, body=resp_data
        )

    def get_headers(
        self, session: 'ks_session.Session'
    ) -> ty.Optional[dict[str, str]]:
        """Fetch authentication headers for message.

        :param session: The session object that the auth_plugin belongs to.
        :type session: keystoneauth1.session.Session

        :returns: Headers that are set to authenticate a message or None for
                  failure. Note that when checking this value that the empty
                  dict is a valid, non-failure response.
        :rtype: dict
        """
        # get headers for X-Auth-Token
        headers = super().get_headers(session)

        # add OAuth2.0 access token to the headers
        if headers:
            headers['Authorization'] = f'Bearer {self.oauth2_access_token}'
        else:
            headers = {'Authorization': f'Bearer {self.oauth2_access_token}'}
        return headers



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty

from keystoneauth1 import identity
from keystoneauth1 import loading
from keystoneauth1.loading import opts


class Token(loading.BaseGenericLoader[identity.Token]):
    """Given an existing token rescope it to another target.

    Use the Identity service's rescope mechanism to get a new token based upon
    an existing token. Because an auth plugin requires a service catalog and
    scope information it is often easier to fetch a new token based on an
    existing one than validate and reuse the one you already have.

    As a generic plugin this plugin is identity version independent and will
    discover available versions before use. This means it expects to be
    provided an unversioned URL to operate against.
    """

    @property
    def plugin_class(self) -> ty.Type[identity.Token]:
        return identity.Token

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()

        options.extend(
            [
                loading.Opt(
                    'token', secret=True, help='Token to authenticate with'
                )
            ]
        )

        return options


class Password(loading.BaseGenericLoader[identity.Password]):
    """Authenticate with a username and password.

    Authenticate to the identity service using the provided username and
    password. This is the standard and most common form of authentication.

    As a generic plugin this plugin is identity version independent and will
    discover available versions before use. This means it expects to be
    provided an unversioned URL to operate against.
    """

    @property
    def plugin_class(self) -> ty.Type[identity.Password]:
        return identity.Password

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()
        options.extend(
            [
                loading.Opt('user-id', help='User id'),
                loading.Opt(
                    'username',
                    help='Username',
                    deprecated=[loading.Opt('user-name')],
                ),
                loading.Opt('user-domain-id', help="User's domain id"),
                loading.Opt('user-domain-name', help="User's domain name"),
                loading.Opt(
                    'password',
                    secret=True,
                    prompt='Password: ',
                    help="User's password",
                ),
            ]
        )
        return options



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import textwrap
from unittest import mock

import pycodestyle
import testtools

from keystoneauth1.hacking import checks
from keystoneauth1.tests.unit import keystoneauth_fixtures


class TestCheckOsloNamespaceImports(testtools.TestCase):
    # We are patching pycodestyle so that only the check under test is actually
    # installed.
    @mock.patch(
        'pycodestyle._checks',
        {'physical_line': {}, 'logical_line': {}, 'tree': {}},
    )
    def run_check(self, code):
        pycodestyle.register_check(checks.check_oslo_namespace_imports)

        lines = textwrap.dedent(code).strip().splitlines(True)

        checker = pycodestyle.Checker(lines=lines)
        checker.check_all()
        checker.report._deferred_print.sort()
        return checker.report._deferred_print

    def assert_has_errors(self, code, expected_errors=None):
        actual_errors = [e[:3] for e in self.run_check(code)]
        self.assertEqual(expected_errors or [], actual_errors)

    def test(self):
        code_ex = self.useFixture(keystoneauth_fixtures.HackingCode())
        code = code_ex.oslo_namespace_imports['code']
        errors = code_ex.oslo_namespace_imports['expected_errors']
        self.assert_has_errors(code, expected_errors=errors)



# Copyright 2022 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty

import requests.auth

from keystoneauth1.exceptions import ClientException
from keystoneauth1.identity.v3 import base
from keystoneauth1 import session as ks_session

__all__ = ('OAuth2ClientCredentialMethod', 'OAuth2ClientCredential')


class OAuth2ClientCredentialMethod(base.AuthMethod):
    """An auth method to fetch a token via an OAuth2.0 client credential.

    :param string oauth2_endpoint: OAuth2.0 endpoint.
    :param string oauth2_client_id: OAuth2.0 client credential id.
    :param string oauth2_client_secret: OAuth2.0 client credential secret.
    """

    oauth2_endpoint: str
    oauth2_client_id: str
    oauth2_client_secret: str

    def __init__(
        self,
        *,
        oauth2_endpoint: str,
        oauth2_client_id: str,
        oauth2_client_secret: str,
    ) -> None:
        self.oauth2_endpoint = oauth2_endpoint
        self.oauth2_client_id = oauth2_client_id
        self.oauth2_client_secret = oauth2_client_secret

    def get_auth_data(
        self,
        session: ks_session.Session,
        auth: base.Auth,
        headers: dict[str, str],
        request_kwargs: dict[str, object],
    ) -> ty.Union[tuple[None, None], tuple[str, ty.Mapping[str, object]]]:
        """Return the authentication section of an auth plugin.

        :param session: The communication session.
        :type session: keystoneauth1.session.Session
        :param base.Auth auth: The auth plugin calling the method.
        :param dict headers: The headers that will be sent with the auth
                             request if a plugin needs to add to them.
        :return: The identifier of this plugin and a dict of authentication
                 data for the auth type.
        :rtype: tuple(string, dict)
        """
        auth_data: dict[str, object] = {
            'id': self.oauth2_client_id,
            'secret': self.oauth2_client_secret,
        }
        return 'application_credential', auth_data

    def get_cache_id_elements(self) -> dict[str, ty.Optional[str]]:
        """Get the elements for this auth method that make it unique.

        These elements will be used as part of the
        :py:meth:`keystoneauth1.plugin.BaseIdentityPlugin.get_cache_id` to
        allow caching of the auth plugin.

        Plugins should override this if they want to allow caching of their
        state.

        To avoid collision or overrides the keys of the returned dictionary
        should be prefixed with the plugin identifier. For example the password
        plugin returns its username value as 'password_username'.
        """
        return {
            'oauth2_client_credential_oauth2_endpoint': self.oauth2_endpoint,
            'oauth2_client_credential_oauth2_client_id': self.oauth2_client_id,
            'oauth2_client_credential_oauth2_client_secret': self.oauth2_client_secret,
        }


class OAuth2ClientCredential(base.Auth):
    """A plugin for authenticating via an OAuth2.0 client credential.

    :param string auth_url: Identity service endpoint for authentication.
    :param string oauth2_endpoint: OAuth2.0 endpoint.
    :param string oauth2_client_id: OAuth2.0 client credential id.
    :param string oauth2_client_secret: OAuth2.0 client credential secret.
    """

    _auth_method_class = OAuth2ClientCredentialMethod

    def __init__(
        self,
        auth_url: str,
        oauth2_endpoint: str,
        oauth2_client_id: str,
        oauth2_client_secret: str,
        *,
        trust_id: ty.Optional[str] = None,
        system_scope: ty.Optional[str] = None,
        domain_id: ty.Optional[str] = None,
        domain_name: ty.Optional[str] = None,
        project_id: ty.Optional[str] = None,
        project_name: ty.Optional[str] = None,
        project_domain_id: ty.Optional[str] = None,
        project_domain_name: ty.Optional[str] = None,
        reauthenticate: bool = True,
        include_catalog: bool = True,
    ) -> None:
        method = self._auth_method_class(
            oauth2_endpoint=oauth2_endpoint,
            oauth2_client_id=oauth2_client_id,
            oauth2_client_secret=oauth2_client_secret,
        )
        super().__init__(
            auth_url,
            [method],
            trust_id=trust_id,
            system_scope=system_scope,
            domain_id=domain_id,
            domain_name=domain_name,
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
            reauthenticate=reauthenticate,
            include_catalog=include_catalog,
        )
        self.oauth2_endpoint = oauth2_endpoint
        self.oauth2_client_id = oauth2_client_id
        self.oauth2_client_secret = oauth2_client_secret

    def get_headers(
        self, session: 'ks_session.Session'
    ) -> ty.Optional[dict[str, str]]:
        """Fetch authentication headers for message.

        :param session: The session object that the auth_plugin belongs to.
        :type session: keystoneauth1.session.Session

        :returns: Headers that are set to authenticate a message or None for
                  failure. Note that when checking this value that the empty
                  dict is a valid, non-failure response.
        :rtype: dict
        """
        # get headers for X-Auth-Token
        headers = super().get_headers(session)

        # Get OAuth2.0 access token and add the field 'Authorization'
        data = {"grant_type": "client_credentials"}
        auth = requests.auth.HTTPBasicAuth(
            self.oauth2_client_id, self.oauth2_client_secret
        )
        resp = session.request(
            self.oauth2_endpoint,
            "POST",
            authenticated=False,
            raise_exc=False,
            data=data,
            requests_auth=auth,
        )
        if resp.status_code == 200:
            oauth2 = resp.json()
            oauth2_token = oauth2["access_token"]
            if headers:
                headers['Authorization'] = f'Bearer {oauth2_token}'
            else:
                headers = {'Authorization': f'Bearer {oauth2_token}'}
        else:
            error = resp.json()
            msg = error.get("error_description")
            raise ClientException(msg)

        return headers



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty

import requests

from keystoneauth1 import access
from keystoneauth1 import exceptions
from keystoneauth1.identity import base
from keystoneauth1.identity.v3 import federation
from keystoneauth1 import plugin
from keystoneauth1 import session as ks_session

__all__ = ('Keystone2Keystone',)


class Keystone2Keystone(federation._Rescoped):
    """Plugin to execute the Keystone to Keyestone authentication flow.

    In this plugin, an ECP wrapped SAML assertion provided by a keystone
    Identity Provider (IdP) is used to request an OpenStack unscoped token
    from a keystone Service Provider (SP).

    :param base_plugin: Auth plugin already authenticated against the keystone
                        IdP.
    :type base_plugin: keystoneauth1.identity.v3.base.BaseAuth

    :param service_provider: The Service Provider ID as returned by
                             ServiceProviderManager.list()
    :type service_provider: str

    """

    REQUEST_ECP_URL = '/auth/OS-FEDERATION/saml2/ecp'
    """Path where the ECP wrapped SAML assertion should be presented to the
       Keystone Service Provider."""

    HTTP_MOVED_TEMPORARILY = 302
    HTTP_SEE_OTHER = 303

    def __init__(
        self,
        base_plugin: base.BaseIdentityPlugin,
        service_provider: str,
        *,
        trust_id: ty.Optional[str] = None,
        system_scope: ty.Optional[str] = None,
        domain_id: ty.Optional[str] = None,
        domain_name: ty.Optional[str] = None,
        project_id: ty.Optional[str] = None,
        project_name: ty.Optional[str] = None,
        project_domain_id: ty.Optional[str] = None,
        project_domain_name: ty.Optional[str] = None,
        reauthenticate: bool = True,
        include_catalog: bool = True,
    ):
        super().__init__(
            auth_url='',
            trust_id=trust_id,
            system_scope=system_scope,
            domain_id=domain_id,
            domain_name=domain_name,
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
            reauthenticate=reauthenticate,
            include_catalog=include_catalog,
        )

        self._local_cloud_plugin = base_plugin
        self._sp_id = service_provider

    @classmethod
    def _remote_auth_url(cls, auth_url: str) -> str:
        """Return auth_url of the remote Keystone Service Provider.

        Remote cloud's auth_url is an endpoint for getting federated unscoped
        token, typically that would be
        ``https://remote.example.com:5000/v3/OS-FEDERATION/identity_providers/
        <idp>/protocols/<protocol_id>/auth``. However we need to generate a
        real auth_url, used for token scoping.  This function assumes there are
        static values today in the remote auth_url stored in the Service
        Provider attribute and those can be used as a delimiter. If the
        sp_auth_url doesn't comply with standard federation auth url the
        function will simply return whole string.

        :param auth_url: auth_url of the remote cloud
        :type auth_url: str

        :returns: auth_url of remote cloud where a token can be validated or
                  scoped.
        :rtype: str

        """
        PATTERN = '/OS-FEDERATION/'
        idx = auth_url.index(PATTERN) if PATTERN in auth_url else len(auth_url)
        return auth_url[:idx]

    def _get_ecp_assertion(self, session: ks_session.Session) -> str:
        body = {
            'auth': {
                'identity': {
                    'methods': ['token'],
                    'token': {
                        'id': self._local_cloud_plugin.get_token(session)
                    },
                },
                'scope': {'service_provider': {'id': self._sp_id}},
            }
        }

        endpoint_filter = {
            'version': (3, 0),
            'interface': plugin.AUTH_INTERFACE,
        }

        headers = {'Accept': 'application/json'}

        resp = session.post(
            self.REQUEST_ECP_URL,
            json=body,
            auth=self._local_cloud_plugin,
            endpoint_filter=endpoint_filter,
            headers=headers,
            authenticated=False,
            raise_exc=False,
        )

        # NOTE(marek-denis): I am not sure whether disabling exceptions in the
        # Session object and testing if resp.ok is sufficient. An alternative
        # would be catching locally all exceptions and reraising with custom
        # warning.
        if not resp.ok:
            msg = (
                "Error while requesting ECP wrapped assertion: response "
                "exit code: %(status_code)d, reason: %(err)s"
            )
            msg = msg % {'status_code': resp.status_code, 'err': resp.reason}
            raise exceptions.AuthorizationFailure(msg)

        if not resp.text:
            raise exceptions.InvalidResponse(resp)

        return str(resp.text)

    def _send_service_provider_ecp_authn_response(
        self, session: ks_session.Session, sp_url: str, sp_auth_url: str
    ) -> requests.Response:
        """Present ECP wrapped SAML assertion to the keystone SP.

        The assertion is issued by the keystone IdP and it is targeted to the
        keystone that will serve as Service Provider.

        :param session: a session object to send out HTTP requests.

        :param sp_url: URL where the ECP wrapped SAML assertion will be
                       presented to the keystone SP. Usually, something like:
                       https://sp.com/Shibboleth.sso/SAML2/ECP
        :type sp_url: str

        :param sp_auth_url: Federated authentication URL of the keystone SP.
                            It is specified by IdP, for example:
                            https://sp.com/v3/OS-FEDERATION/identity_providers/
                            idp_id/protocols/protocol_id/auth
        :type sp_auth_url: str

        """
        response = session.post(
            sp_url,
            headers={'Content-Type': 'application/vnd.paos+xml'},
            data=self._get_ecp_assertion(session),
            authenticated=False,
            redirect=False,
        )

        # Don't follow HTTP specs - after the HTTP 302/303 response don't
        # repeat the call directed to the Location URL. In this case, this is
        # an indication that SAML2 session is now active and protected resource
        # can be accessed.
        if response.status_code in (
            self.HTTP_MOVED_TEMPORARILY,
            self.HTTP_SEE_OTHER,
        ):
            response = session.get(
                sp_auth_url,
                headers={'Content-Type': 'application/vnd.paos+xml'},
                authenticated=False,
            )

        return response

    def get_unscoped_auth_ref(
        self, session: ks_session.Session
    ) -> access.AccessInfoV3:
        sp_auth_url = self._local_cloud_plugin.get_sp_auth_url(
            session, self._sp_id
        )
        sp_url = self._local_cloud_plugin.get_sp_url(session, self._sp_id)
        assert sp_auth_url is not None  # nosec B101
        assert sp_url is not None  # nosec B101
        self.auth_url = self._remote_auth_url(sp_auth_url)

        response = self._send_service_provider_ecp_authn_response(
            session, sp_url, sp_auth_url
        )
        access_info = access.create(resp=response)
        assert isinstance(access_info, access.AccessInfoV3)  # nosec B101
        return access_info



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

UNSCOPED_TOKEN_HEADER = 'UNSCOPED_TOKEN'

UNSCOPED_TOKEN = {
    "token": {
        "issued_at": "2014-06-09T09:48:59.643406Z",
        "extras": {},
        "methods": ["token"],
        "expires_at": "2014-06-09T10:48:59.643375Z",
        "user": {
            "OS-FEDERATION": {
                "identity_provider": {"id": "testshib"},
                "protocol": {"id": "saml2"},
                "groups": [{"id": "1764fa5cf69a49a4918131de5ce4af9a"}],
            },
            "id": "testhib%20user",
            "name": "testhib user",
        },
    }
}


SAML_ENCODING = "<?xml version='1.0' encoding='UTF-8'?>"

TOKEN_SAML_RESPONSE = """
<ns2:Response Destination="http://beta.example.com/Shibboleth.sso/POST/ECP"
  ID="8c21de08d2f2435c9acf13e72c982846"
  IssueInstant="2015-03-25T14:43:21Z"
  Version="2.0">
  <saml:Issuer Format="urn:oasis:names:tc:SAML:2.0:nameid-format:entity">
    http://keystone.idp/v3/OS-FEDERATION/saml2/idp
  </saml:Issuer>
  <ns2:Status>
    <ns2:StatusCode Value="urn:oasis:names:tc:SAML:2.0:status:Success"/>
  </ns2:Status>
  <saml:Assertion ID="a5f02efb0bff4044b294b4583c7dfc5d"
    IssueInstant="2015-03-25T14:43:21Z" Version="2.0">
  <saml:Issuer Format="urn:oasis:names:tc:SAML:2.0:nameid-format:entity">
    http://keystone.idp/v3/OS-FEDERATION/saml2/idp</saml:Issuer>
  <xmldsig:Signature>
    <xmldsig:SignedInfo>
      <xmldsig:CanonicalizationMethod
        Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"/>
      <xmldsig:SignatureMethod
        Algorithm="http://www.w3.org/2000/09/xmldsig#rsa-sha1"/>
      <xmldsig:Reference URI="#a5f02efb0bff4044b294b4583c7dfc5d">
        <xmldsig:Transforms>
          <xmldsig:Transform
             Algorithm="http://www.w3.org/2000/09/xmldsig#
             enveloped-signature"/>
          <xmldsig:Transform
             Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"/>
        </xmldsig:Transforms>
        <xmldsig:DigestMethod
          Algorithm="http://www.w3.org/2000/09/xmldsig#sha1"/>
        <xmldsig:DigestValue>
          0KH2CxdkfzU+6eiRhTC+mbObUKI=
        </xmldsig:DigestValue>
      </xmldsig:Reference>
    </xmldsig:SignedInfo>
    <xmldsig:SignatureValue>
      m2jh5gDvX/1k+4uKtbb08CHp2b9UWsLw
    </xmldsig:SignatureValue>
    <xmldsig:KeyInfo>
      <xmldsig:X509Data>
        <xmldsig:X509Certificate>...</xmldsig:X509Certificate>
      </xmldsig:X509Data>
    </xmldsig:KeyInfo>
  </xmldsig:Signature>
  <saml:Subject>
    <saml:NameID>admin</saml:NameID>
    <saml:SubjectConfirmation Method="urn:oasis:names:tc:SAML:2.0:cm:bearer">
      <saml:SubjectConfirmationData
        NotOnOrAfter="2015-03-25T15:43:21.172385Z"
        Recipient="http://beta.example.com/Shibboleth.sso/POST/ECP"/>
    </saml:SubjectConfirmation>
  </saml:Subject>
  <saml:AuthnStatement AuthnInstant="2015-03-25T14:43:21Z"
    SessionIndex="9790eb729858456f8a33b7a11f0a637e"
    SessionNotOnOrAfter="2015-03-25T15:43:21.172385Z">
    <saml:AuthnContext>
      <saml:AuthnContextClassRef>
        urn:oasis:names:tc:SAML:2.0:ac:classes:Password
      </saml:AuthnContextClassRef>
      <saml:AuthenticatingAuthority>
        http://keystone.idp/v3/OS-FEDERATION/saml2/idp
      </saml:AuthenticatingAuthority>
    </saml:AuthnContext>
  </saml:AuthnStatement>
  <saml:AttributeStatement>
    <saml:Attribute Name="openstack_user"
      NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:uri">
      <saml:AttributeValue xsi:type="xs:string">admin</saml:AttributeValue>
    </saml:Attribute>
    <saml:Attribute Name="openstack_roles"
      NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:uri">
      <saml:AttributeValue xsi:type="xs:string">admin</saml:AttributeValue>
    </saml:Attribute>
    <saml:Attribute Name="openstack_project"
      NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:uri">
      <saml:AttributeValue xsi:type="xs:string">admin</saml:AttributeValue>
    </saml:Attribute>
  </saml:AttributeStatement>
  </saml:Assertion>
</ns2:Response>
"""

TOKEN_BASED_SAML = ''.join([SAML_ENCODING, TOKEN_SAML_RESPONSE])

ECP_ENVELOPE = f"""
<ns0:Envelope
  xmlns:ns0="http://schemas.xmlsoap.org/soap/envelope/"
  xmlns:ns1="urn:oasis:names:tc:SAML:2.0:profiles:SSO:ecp"
  xmlns:ns2="urn:oasis:names:tc:SAML:2.0:protocol"
  xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion"
  xmlns:xmldsig="http://www.w3.org/2000/09/xmldsig#"
  xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <ns0:Header>
    <ns1:RelayState
      ns0:actor="http://schemas.xmlsoap.org/soap/actor/next"
      ns0:mustUnderstand="1">
        ss:mem:1ddfe8b0f58341a5a840d2e8717b0737
      </ns1:RelayState>
  </ns0:Header>
  <ns0:Body>
  {TOKEN_SAML_RESPONSE}
  </ns0:Body>
</ns0:Envelope>
"""

TOKEN_BASED_ECP = ''.join([SAML_ENCODING, ECP_ENVELOPE])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystoneauth1.exceptions import base

__all__ = ('ServiceProviderNotFound',)


class ServiceProviderNotFound(base.ClientException):
    """A Service Provider cannot be found."""

    def __init__(self, sp_id: str):
        self.sp_id = sp_id
        msg = f'The Service Provider {sp_id} could not be found'
        super().__init__(msg)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import requests

from keystoneauth1 import _utils as utils
from keystoneauth1.exceptions import base


class AuthorizationFailure(base.ClientException):
    message = "Cannot authorize API client."


class MissingAuthMethods(base.ClientException):
    message = "Not all required auth rules were satisfied"

    def __init__(self, response: requests.Response):
        self.response = response
        self.receipt = response.headers.get("Openstack-Auth-Receipt")
        body = response.json()
        self.methods = body['receipt']['methods']
        self.required_auth_methods = body['required_auth_methods']
        self.expires_at = utils.parse_isotime(body['receipt']['expires_at'])
        message = f"{self.message}: {self.required_auth_methods}"
        super().__init__(message)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystoneauth1.identity import base
from keystoneauth1.identity import generic
from keystoneauth1.identity import v2
from keystoneauth1.identity import v3
from keystoneauth1.identity.v3 import oidc


BaseIdentityPlugin = base.BaseIdentityPlugin

V2Password = v2.Password
"""See :class:`keystoneauth1.identity.v2.Password`"""

V2Token = v2.Token
"""See :class:`keystoneauth1.identity.v2.Token`"""

V3Password = v3.Password
"""See :class:`keystoneauth1.identity.v3.Password`"""

V3Token = v3.Token
"""See :class:`keystoneauth1.identity.v3.Token`"""

Password = generic.Password
"""See :class:`keystoneauth1.identity.generic.Password`"""

Token = generic.Token
"""See :class:`keystoneauth1.identity.generic.Token`"""

V3OidcClientCredentials = oidc.OidcClientCredentials
"""See :class:`keystoneauth1.identity.v3.oidc.OidcClientCredentials`"""

V3OidcPassword = oidc.OidcPassword
"""See :class:`keystoneauth1.identity.v3.oidc.OidcPassword`"""

V3OidcAuthorizationCode = oidc.OidcAuthorizationCode
"""See :class:`keystoneauth1.identity.v3.oidc.OidcAuthorizationCode`"""

V3OidcAccessToken = oidc.OidcAccessToken
"""See :class:`keystoneauth1.identity.v3.oidc.OidcAccessToken`"""

V3OidcDeviceAuthorization = oidc.OidcDeviceAuthorization
"""See :class:`keystoneauth1.identity.v3.oidc.OidcDeviceAuthorization`"""

V3TOTP = v3.TOTP
"""See :class:`keystoneauth1.identity.v3.TOTP`"""

V3TokenlessAuth = v3.TokenlessAuth
"""See :class:`keystoneauth1.identity.v3.TokenlessAuth`"""

V3ApplicationCredential = v3.ApplicationCredential
"""See :class:`keystoneauth1.identity.v3.ApplicationCredential`"""

V3MultiFactor = v3.MultiFactor
"""See :class:`keystoneauth1.identity.v3.MultiFactor`"""

V3OAuth2ClientCredential = v3.OAuth2ClientCredential
"""See :class:`keystoneauth1.identity.v3.OAuth2ClientCredential`"""

V3OAuth2mTlsClientCredential = v3.OAuth2mTlsClientCredential
"""See :class:`keystoneauth1.identity.v3.OAuth2mTlsClientCredential`"""

__all__ = (
    'BaseIdentityPlugin',
    'Password',
    'Token',
    'V2Password',
    'V2Token',
    'V3Password',
    'V3Token',
    'V3OidcPassword',
    'V3OidcAuthorizationCode',
    'V3OidcAccessToken',
    'V3OidcDeviceAuthorization',
    'V3TOTP',
    'V3TokenlessAuth',
    'V3ApplicationCredential',
    'V3MultiFactor',
    'V3OAuth2ClientCredential',
    'V3OAuth2mTlsClientCredential',
)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty

from keystoneauth1.identity.v3 import base
from keystoneauth1 import session as ks_session

__all__ = ('PasswordMethod', 'Password')


class PasswordMethod(base.AuthMethod):
    """Construct a User/Password based authentication method.

    :param string password: Password for authentication.
    :param string username: Username for authentication.
    :param string user_id: User ID for authentication.
    :param string user_domain_id: User's domain ID for authentication.
    :param string user_domain_name: User's domain name for authentication.
    """

    password: str
    user_id: ty.Optional[str] = None
    username: ty.Optional[str] = None
    user_domain_id: ty.Optional[str] = None
    user_domain_name: ty.Optional[str] = None

    def __init__(
        self,
        *,
        password: str,
        user_id: ty.Optional[str] = None,
        username: ty.Optional[str] = None,
        user_domain_id: ty.Optional[str] = None,
        user_domain_name: ty.Optional[str] = None,
    ) -> None:
        self.password = password
        self.user_id = user_id
        self.username = username
        self.user_domain_id = user_domain_id
        self.user_domain_name = user_domain_name

    def get_auth_data(
        self,
        session: ks_session.Session,
        auth: base.Auth,
        headers: dict[str, str],
        request_kwargs: dict[str, object],
    ) -> ty.Union[tuple[None, None], tuple[str, ty.Mapping[str, object]]]:
        user: dict[str, ty.Any] = {'password': self.password}

        if self.user_id:
            user['id'] = self.user_id
        elif self.username:
            user['name'] = self.username

            if self.user_domain_id:
                user['domain'] = {'id': self.user_domain_id}
            elif self.user_domain_name:
                user['domain'] = {'name': self.user_domain_name}

        return 'password', {'user': user}

    def get_cache_id_elements(self) -> dict[str, ty.Optional[str]]:
        return {
            'password_password': self.password,
            'password_user_id': self.user_id,
            'password_username': self.username,
            'password_user_domain_id': self.user_domain_id,
            'password_user_domain_name': self.user_domain_name,
        }


class Password(base.Auth):
    """A plugin for authenticating with a username and password.

    :param string auth_url: Identity service endpoint for authentication.
    :param string password: Password for authentication.
    :param string user_id: User ID for authentication.
    :param string username: Username for authentication.
    :param string user_domain_id: User's domain ID for authentication.
    :param string user_domain_name: User's domain name for authentication.
    :param string trust_id: Trust ID for trust scoping.
    :param string system_scope: System information to scope to.
    :param string domain_id: Domain ID for domain scoping.
    :param string domain_name: Domain name for domain scoping.
    :param string project_id: Project ID for project scoping.
    :param string project_name: Project name for project scoping.
    :param string project_domain_id: Project's domain ID for project.
    :param string project_domain_name: Project's domain name for project.
    :param bool reauthenticate: Allow fetching a new token if the current one
                                is going to expire. (optional) default True
    """

    _auth_method_class = PasswordMethod

    def __init__(
        self,
        auth_url: str,
        password: str,
        user_id: ty.Optional[str] = None,
        username: ty.Optional[str] = None,
        user_domain_id: ty.Optional[str] = None,
        user_domain_name: ty.Optional[str] = None,
        *,
        unscoped: bool = False,
        trust_id: ty.Optional[str] = None,
        system_scope: ty.Optional[str] = None,
        domain_id: ty.Optional[str] = None,
        domain_name: ty.Optional[str] = None,
        project_id: ty.Optional[str] = None,
        project_name: ty.Optional[str] = None,
        project_domain_id: ty.Optional[str] = None,
        project_domain_name: ty.Optional[str] = None,
        reauthenticate: bool = True,
        include_catalog: bool = True,
    ) -> None:
        method = self._auth_method_class(
            password=password,
            user_id=user_id,
            username=username,
            user_domain_id=user_domain_id,
            user_domain_name=user_domain_name,
        )
        super().__init__(
            auth_url,
            [method],
            unscoped=unscoped,
            trust_id=trust_id,
            system_scope=system_scope,
            domain_id=domain_id,
            domain_name=domain_name,
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
            reauthenticate=reauthenticate,
            include_catalog=include_catalog,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import typing as ty

from keystoneauth1 import adapter
from keystoneauth1.loading import _utils
from keystoneauth1.loading import base

if ty.TYPE_CHECKING:
    from oslo_config import cfg

    from keystoneauth1.loading import opts


__all__ = (
    'register_argparse_arguments',
    'register_service_argparse_arguments',
    'register_conf_options',
    'load_from_conf_options',
    'get_conf_options',
)


class Adapter(base._BaseLoader[adapter.Adapter]):
    @property
    def plugin_class(self) -> ty.Type[adapter.Adapter]:
        return adapter.Adapter

    def get_options(self) -> list['opts.Opt']:
        return []

    @staticmethod
    def get_conf_options(
        include_deprecated: bool = True,
        deprecated_opts: ty.Optional[
            dict[str, list['cfg.DeprecatedOpt']]
        ] = None,
    ) -> list['cfg.Opt']:
        """Get oslo_config options that are needed for a :py:class:`.Adapter`.

        These may be useful without being registered for config file generation
        or to manipulate the options before registering them yourself.

        The options that are set are:
            :service_type:      The default service_type for URL discovery.
            :service_name:      The default service_name for URL discovery.
            :interface:         The default interface for URL discovery.
                                (deprecated)
            :valid_interfaces:  List of acceptable interfaces for URL
                                discovery. Can be a list of any of
                                'public', 'internal' or 'admin'.
            :region_name:       The default region_name for URL discovery.
            :endpoint_override: Always use this endpoint URL for requests
                                for this client.
            :version:           The minimum version restricted to a given Major
                                API. Mutually exclusive with min_version and
                                max_version.
            :min_version:       The minimum major version of a given API,
                                intended to be used as the lower bound of a
                                range with max_version. Mutually exclusive with
                                version. If min_version is given with no
                                max_version it is as if max version is
                                'latest'.
            :max_version:       The maximum major version of a given API,
                                intended to be used as the upper bound of a
                                range with min_version. Mutually exclusive with
                                version.

        :param include_deprecated: If True (the default, for backward
                                   compatibility), deprecated options are
                                   included in the result.  If False, they are
                                   excluded.
        :param dict deprecated_opts: Deprecated options that should be included
             in the definition of new options. This should be a dict from the
             name of the new option to a list of oslo.DeprecatedOpts that
             correspond to the new option. (optional)

             For example, to support the ``api_endpoint`` option pointing to
             the new ``endpoint_override`` option name::

                 old_opt = oslo_cfg.DeprecatedOpt('api_endpoint', 'old_group')
                 deprecated_opts = {'endpoint_override': [old_opt]}

        :returns: A list of oslo_config options.
        """
        cfg = _utils.get_oslo_config()

        if deprecated_opts is None:
            deprecated_opts = {}

        # This is goofy, but need to support hyphens *or* underscores
        deprecated_opts = {
            name.replace('_', '-'): opt
            for name, opt in deprecated_opts.items()
        }

        opts = [
            cfg.StrOpt(
                'service-type',
                deprecated_opts=deprecated_opts.get('service-type'),
                help='The default service_type for endpoint URL discovery.',
            ),
            cfg.StrOpt(
                'service-name',
                deprecated_opts=deprecated_opts.get('service-name'),
                help='The default service_name for endpoint URL discovery.',
            ),
            cfg.ListOpt(
                'valid-interfaces',
                deprecated_opts=deprecated_opts.get('valid-interfaces'),
                help='List of interfaces, in order of preference, '
                'for endpoint URL.',
            ),
            cfg.StrOpt(
                'region-name',
                deprecated_opts=deprecated_opts.get('region-name'),
                help='The default region_name for endpoint URL discovery.',
            ),
            cfg.StrOpt(
                'endpoint-override',
                deprecated_opts=deprecated_opts.get('endpoint-override'),
                help='Always use this endpoint URL for requests '
                'for this client. NOTE: The unversioned '
                'endpoint should be specified here; to '
                'request a particular API version, use the '
                '`version`, `min-version`, and/or '
                '`max-version` options.',
            ),
            cfg.StrOpt(
                'version',
                deprecated_opts=deprecated_opts.get('version'),
                help='Minimum Major API version within a given '
                'Major API version for endpoint URL '
                'discovery. Mutually exclusive with '
                'min_version and max_version',
            ),
            cfg.StrOpt(
                'min-version',
                deprecated_opts=deprecated_opts.get('min-version'),
                help='The minimum major version of a given API, '
                'intended to be used as the lower bound of a '
                'range with max_version. Mutually exclusive '
                'with version. If min_version is given with '
                'no max_version it is as if max version is '
                '"latest".',
            ),
            cfg.StrOpt(
                'max-version',
                deprecated_opts=deprecated_opts.get('max-version'),
                help='The maximum major version of a given API, '
                'intended to be used as the upper bound of a '
                'range with min_version. Mutually exclusive '
                'with version.',
            ),
            cfg.IntOpt(
                'connect-retries',
                deprecated_opts=deprecated_opts.get('connect-retries'),
                help='The maximum number of retries that should be '
                'attempted for connection errors.',
            ),
            cfg.FloatOpt(
                'connect-retry-delay',
                deprecated_opts=deprecated_opts.get('connect-retry-delay'),
                help='Delay (in seconds) between two retries '
                'for connection errors. If not set, '
                'exponential retry starting with 0.5 '
                'seconds up to a maximum of 60 seconds '
                'is used.',
            ),
            cfg.IntOpt(
                'status-code-retries',
                deprecated_opts=deprecated_opts.get('status-code-retries'),
                help='The maximum number of retries that should be '
                'attempted for retriable HTTP status codes.',
            ),
            cfg.FloatOpt(
                'status-code-retry-delay',
                deprecated_opts=deprecated_opts.get('status-code-retry-delay'),
                help='Delay (in seconds) between two retries '
                'for retriable status codes. If not set, '
                'exponential retry starting with 0.5 '
                'seconds up to a maximum of 60 seconds '
                'is used.',
            ),
            cfg.ListOpt(
                'retriable-status-codes',
                deprecated_opts=deprecated_opts.get('retriable-status-codes'),
                item_type=cfg.types.Integer(),
                help='List of retriable HTTP status codes that '
                'should be retried. If not set default to '
                ' [503]',
            ),
        ]

        if include_deprecated:
            opts += [
                cfg.StrOpt(
                    'interface',
                    help='The default interface for endpoint URL discovery.',
                    deprecated_for_removal=True,
                    deprecated_reason='Using valid-interfaces is'
                    ' preferrable because it is'
                    ' capable of accepting a list of'
                    ' possible interfaces.',
                )
            ]
        return opts

    def register_conf_options(
        self,
        conf: 'cfg.ConfigOpts',
        group: str,
        include_deprecated: bool = True,
        deprecated_opts: ty.Optional[
            dict[str, list['cfg.DeprecatedOpt']]
        ] = None,
    ) -> list['cfg.Opt']:
        """Register the oslo_config options that are needed for an Adapter.

        The options that are set are:
            :service_type:        The default service_type for URL discovery.
            :service_name:        The default service_name for URL discovery.
            :interface:           The default interface for URL discovery.
                                  (deprecated)
            :valid_interfaces:    List of acceptable interfaces for URL
                                  discovery. Can be a list of any of
                                  'public', 'internal' or 'admin'.
            :region_name:         The default region_name for URL discovery.
            :endpoint_override:   Always use this endpoint URL for requests
                                  for this client.
            :version:             The minimum version restricted to a given
                                  Major API. Mutually exclusive with
                                  min_version and max_version.
            :min_version:         The minimum major version of a given API,
                                  intended to be used as the lower bound of a
                                  range with max_version. Mutually exclusive
                                  with version. If min_version is given with no
                                  max_version it is as if max version is
                                  'latest'.
            :max_version:         The maximum major version of a given API,
                                  intended to be used as the upper bound of a
                                  range with min_version. Mutually exclusive
                                  with version.
            :connect_retries:     The maximum number of retries that should be
                                  attempted for connection errors.
            :status_code_retries: The maximum number of retries that should be
                                  attempted for retriable HTTP status codes.

        :param oslo_config.Cfg conf: config object to register with.
        :param string group: The ini group to register options in.
        :param include_deprecated: If True (the default, for backward
                                   compatibility), deprecated options are
                                   registered.  If False, they are excluded.
        :param dict deprecated_opts: Deprecated options that should be included
             in the definition of new options. This should be a dict from the
             name of the new option to a list of oslo.DeprecatedOpts that
             correspond to the new option. (optional)

             For example, to support the ``api_endpoint`` option pointing to
             the new ``endpoint_override`` option name::

                 old_opt = oslo_cfg.DeprecatedOpt('api_endpoint', 'old_group')
                 deprecated_opts = {'endpoint_override': [old_opt]}

        :returns: The list of options that was registered.
        """
        opts = self.get_conf_options(
            include_deprecated=include_deprecated,
            deprecated_opts=deprecated_opts,
        )
        conf.register_group(_utils.get_oslo_config().OptGroup(group))
        conf.register_opts(opts, group=group)
        return opts

    def load_from_conf_options(
        self, conf: 'cfg.ConfigOpts', group: str, **kwargs: ty.Any
    ) -> adapter.Adapter:
        """Create an Adapter object from an oslo_config object.

        The options must have been previously registered with
        register_conf_options.

        :param oslo_config.Cfg conf: config object to register with.
        :param string group: The ini group to register options in.
        :param dict kwargs: Additional parameters to pass to Adapter
                            construction.
        :returns: A new Adapter object.
        :rtype: :py:class:`.Adapter`
        """
        c = conf[group]
        process_conf_options(c, kwargs)
        return self.load_from_options(**kwargs)


def process_conf_options(
    confgrp: 'cfg.OptGroup', kwargs: dict[str, ty.Any]
) -> None:
    """Set Adapter constructor kwargs based on conf options.

    :param oslo_config.cfg.OptGroup confgrp: Config object group containing
            options to inspect.
    :param dict kwargs: Keyword arguments suitable for the constructor of
            keystoneauth1.adapter.Adapter. Will be modified by this method.
            Values already set remain unaffected.
    :raise TypeError: If invalid conf option values or combinations are found.
    """
    if confgrp.valid_interfaces and getattr(confgrp, 'interface', None):
        raise TypeError(
            "interface and valid_interfaces are mutually "
            "exclusive. Please use valid_interfaces."
        )
    if confgrp.valid_interfaces:
        for iface in confgrp.valid_interfaces:
            if iface not in ('public', 'internal', 'admin'):
                raise TypeError(
                    f"'{iface}' is not a valid value for"
                    " valid_interfaces. Valid values are"
                    " public, internal or admin"
                )
        kwargs.setdefault('interface', confgrp.valid_interfaces)
    elif hasattr(confgrp, 'interface'):
        kwargs.setdefault('interface', confgrp.interface)
    kwargs.setdefault('service_type', confgrp.service_type)
    kwargs.setdefault('service_name', confgrp.service_name)
    kwargs.setdefault('region_name', confgrp.region_name)
    kwargs.setdefault('endpoint_override', confgrp.endpoint_override)
    kwargs.setdefault('version', confgrp.version)
    kwargs.setdefault('min_version', confgrp.min_version)
    kwargs.setdefault('max_version', confgrp.max_version)
    if kwargs['version'] and (kwargs['max_version'] or kwargs['min_version']):
        raise TypeError(
            "version is mutually exclusive with min_version and max_version"
        )
    kwargs.setdefault('connect_retries', confgrp.connect_retries)
    kwargs.setdefault('connect_retry_delay', confgrp.connect_retry_delay)
    kwargs.setdefault('status_code_retries', confgrp.status_code_retries)
    kwargs.setdefault(
        'status_code_retry_delay', confgrp.status_code_retry_delay
    )
    kwargs.setdefault('retriable_status_codes', confgrp.retriable_status_codes)


def register_argparse_arguments(
    parser: argparse.ArgumentParser, service_type: ty.Optional[str] = None
) -> None:
    return adapter.register_adapter_argparse_arguments(parser, service_type)


def register_service_argparse_arguments(
    parser: argparse.ArgumentParser, service_type: str
) -> None:
    return adapter.register_service_adapter_argparse_arguments(
        parser, service_type
    )


def register_conf_options(
    conf: 'cfg.ConfigOpts',
    group: str,
    include_deprecated: bool = True,
    deprecated_opts: ty.Optional[dict[str, list['cfg.DeprecatedOpt']]] = None,
) -> list['cfg.Opt']:
    return Adapter().register_conf_options(
        conf,
        group,
        include_deprecated=include_deprecated,
        deprecated_opts=deprecated_opts,
    )


def load_from_conf_options(
    conf: 'cfg.ConfigOpts', group: str, **kwargs: ty.Any
) -> adapter.Adapter:
    return Adapter().load_from_conf_options(conf, group, **kwargs)


def get_conf_options(
    include_deprecated: bool = True,
    deprecated_opts: ty.Optional[dict[str, list['cfg.DeprecatedOpt']]] = None,
) -> list['cfg.Opt']:
    return Adapter.get_conf_options(include_deprecated, deprecated_opts)
